"""
Geonomics Help (:mod:`geonomics.help`)
======================================

.. currentmodule:: geonomics.help

Memory Help
-----------

    get_obj_size    Get the full size of a complex object

Parameters Help
---------------

    plot_movement   Plot example movement tracks generated by the input
                    movement parameters


Background information
----------------------

Geonomics models can occupy a lot of memory, and also contain a large number of
parameters to be customized. The functions in this module aim to provide some
assistance in assessing Geonomics' behavior, so as to help users better plan,
introspect, and debug their models.

"""
from ._param_help import *
from ._memory_help import *



#!/usr/bin/python
# genome.py

# flake8: noqa


'''
Classes, associated methods, and supporting functions for all genomic
components
'''

# geonomics imports
from geonomics.ops import mutation
from geonomics.utils.viz import _check_display

# other imports
import numpy as np
import pandas as pd
from numpy import random as r
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from collections import OrderedDict as OD
import warnings
import random
import bisect
import bitarray
import tskit

######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################


# an error to raise if a simulation is not given enough neutral loci to fit
# the total number of expected mutations over the total parameterized runtime
# under the infinite sites model
class MutationRateError(Exception):
    pass


class Recombinations:
    """
    Container for the parameters and data necessary to simulate recombination.

    Not intended for public use.
    """
    def __init__(self, L, positions, n, r_distr_alpha, r_distr_beta,
                 recomb_rates, jitter_breakpoints):
        # define the bitarray unit corresponding to each homologue
        self._bitarray_dict= {0: '10', 1: '01'}

        # genome length
        self._L = L
        # organize the potential recombination breakpoint positions
        if positions is None:
            positions = np.arange(self._L)
        else:
            positions.sort()
        positions = [*positions]
        # optimize the data-type, to save some memory
        if len(positions) <= 2**16:
            self._positions = np.int16(np.array(positions))
        else:
            self._positions = np.int32(np.array(positions))

        # number of recombination events to simulate and cache for the model
        # NOTE: the higher this number, the more accurately geonomics
        #       will simulate the stipulated recombination rates
        self._n = n
        # alpha and beta parameters for the beta distribution from which
        # recombination rates can be drawn
        self._r_distr_alpha = r_distr_alpha
        self._r_distr_beta = r_distr_beta
        # flag indicating whether or not breakpoint locations should be
        # jittered slightly off of their x.5 defaults, so that
        # tskit.TreeSequence will correctly track and report the number of trees
        self._jitter_breakpoints = jitter_breakpoints
        # get the recombination rates
        if recomb_rates is not None:
            assert len(recomb_rates) == len(positions), ("Lengths of provided "
                                                         "recombination "
                                                         "rates and "
                                                         "recombination "
                                                         "positions don't "
                                                         "match!")
            assert recomb_rates[0] == 0, ("The first recombination rate (i.e. "
                                          "index 0) must be 0 (because no "
                                          "recombination can be modeled as "
                                          "taking place prior to the first "
                                          "locus being simulated).")
            self._rates = recomb_rates
        else:
            self._rates = self._draw_recombination_rates()

    def _set_events(self, use_subsetters, nonneutral_loci):
        # set the potential recombination breakpoints, their recombination
        # rates, and the cache of simulated recombination events to be used
        # by the model
        (self._breakpoints,
         self._subsetters) = self._draw_recombination_events(
                use_subsetters=use_subsetters, nonneutral_loci=nonneutral_loci)
        self._set_seg_info()

    def _get_events(self, size):
        events = random.sample(self._events, size)
        return events

    def _get_subsetter(self, event_key):
        return self._subsetters[event_key]


    # update all the recombination subsetters in accord with a new non-neutral
    # mutation
    def _update_subsetters(self, mutation_loc, genotype_arr_mutation_idx):
        for k in self._subsetters.keys():
            # determine whether the number of crossovers that has happened up
            # to the mutation's locus has been even (in which case the
            # recomb path is 'back' on homologue 0) or odd (in which
            # case it is on homologue 1)
            # NOTE: this is because all recomb paths start on homologue 0,
            #       then are used either for subsetting starting from either
            #       homol 0 or homol 1 on the fly in mating.py using the 
            #       variable start_homologues
            #       in function _do_mating_singl_offspring
            subsetter_homologue = bisect.bisect_left(self._breakpoints[k],
                                                     mutation_loc) % 2
            # use the bitarray_dict to get the 2-digit binary string
            # corresponding to the homologue that the recomb path is on
            # when it reaches mutation_loc, to be inserted into the subsetter
            # in the appropriate location
            subsetter_insert = self._bitarray_dict[subsetter_homologue]
            # find the insertion-point index of the current subsetter 
            # (just twice the non-neutral genotype array's index)
            subsetter_idx = genotype_arr_mutation_idx * 2
            # create the new subsetter by concatenating:
            #  1. the original subsetter, up to but excluding the insertion pt
            #  2. the insert
            #  3. the remainder of the original subsetter
            new_subsetter = (self._subsetters[k][:subsetter_idx] +
                subsetter_insert + self._subsetters[k][subsetter_idx:])
            self._subsetters[k] = new_subsetter


    # draw recombination rates as needed according to parameter values
    def _draw_recombination_rates(self):
        if (self._r_distr_alpha is not None and self._r_distr_beta is not None):
            recomb_rates = np.clip(np.random.beta(a=self._r_distr_alpha,
                                                  b=self._r_distr_beta,
                                                  size=len(self._positions)),
                                   a_min=0, a_max=0.5)
        elif self._r_distr_alpha is not None:
            recomb_rates = np.ones(len(self._positions)) * self._r_distr_alpha
        else:
            #NOTE: if no alpha and beta values were provided for the recomb-rate
            # beta distribution, set all interlocus recomb rates to a value that
            # results in an average of one recombination per gamete produced
            homog_recomb_rate = 1 / self._L
            recomb_rates = np.ones(len(self._positions)) * homog_recomb_rate
        # set the 0th rate to 0 (because this will ensure that all recombination
        # paths start on homologue 0, such that when the start homologue of a
        # certain gamete equals 1 and its genome array is L-R flipped before
        # subsetting, the effective start homologue of the new, subsetted gamete
        # will actually be 1)
        recomb_rates[0] = 0
        return recomb_rates


    # simulate recombination events
    def _draw_recombination_events(self, use_subsetters, nonneutral_loci):
        """
        NOTE: Positions and recomb_rates must be provided already sorted!
        """
        # create the breakpoints dict
        recombinations = [r.binomial(1, self._rates) for _ in range(self._n)]
        breakpoints = [self._positions[np.where(
                                        recomb)] for recomb in recombinations]
        # recast it as an int-keyed dict, so that when I use the recombination
        # events to simulate recombination I don't have to pass around the long
        # arrays, but instead can just random keys to index them out on the fly
        breakpoints = dict(zip(range(len(breakpoints)), breakpoints))

        if use_subsetters:
            # get the recombination paths used to make the bitarray subsetters
            if len(nonneutral_loci) > 0:
                recomb_paths = [np.cumsum(
                                    recomb) % 2 for recomb in recombinations]
                # grab just the values at the nonneutral loci
                # (because the subsetters will be used only for subsetting
                # the non-neutral genotypes carried with the individuals)
                paths = [path[nonneutral_loci] for path in recomb_paths]
                subsetters = [bitarray.bitarray(''.join([self._bitarray_dict[
                            step] for step in path])) for path in paths]
                # recast as integer-indexed dict
            else:
                subsetters = [bitarray.bitarray() for _ in range(
                                                        len(recombinations))]
            subsetters = dict(zip(range(len(subsetters)), subsetters))
            assert len(breakpoints) == len(subsetters)
        else:
            subsetters = None
        return (breakpoints, subsetters)


    # calculate the left and right segment edges for each recombination event
    def _set_seg_info(self):
        seg_info = {}
        for k, recomb in self._breakpoints.items():
            # NOTE: subtracting 0.5 from all recombination breakpoints,
            # to indicate that recombination 'actually' happens halfway between
            # a pair of loci, (i.e. it actually subsets Individuals'
            # genomes in that way) without having the hold all the 0.5's
            # in the Recombinations._events data struct (to save on memory)
            left = np.array([0] + [*recomb - 0.5])
            right = np.array([*recomb - 0.5] + [self._L])
            seg_info[k] = (left, right)
        self._seg_info = seg_info


   # take a recombination event key and a parent's node ids,
   # return a zip object containing 
        # 1.) node id for parent's id (corresponding to the id in the
        # tskit.TableCollection.nodes table);
        # 2.) left end of the segment (inclusive, according to tskit)
        # 3.) right end (exclusive)
    def _get_seg_info(self, start_homologue, event_key,
                                    node_ids):
        left, right = self._seg_info[event_key]
        # add small jitter to the numbers, so that recombination breakpoints
        # are essentially not repeated, and this way tskit can correctly track
        # the number of breakpoints and number of trees
        if self._jitter_breakpoints:
            left, right = [np.clip(seg_bps + np.random.uniform(0, 0.0001,
                                                               len(seg_bps)),
                                   0, self._L) for seg_bps in [left, right]]
        homologue_nodes = node_ids[[(i + start_homologue) % 2 for i in range(
                                                                   len(left))]]
        seg_info = zip(homologue_nodes, left, right)
        return seg_info


class Trait:
    """
    Representation of a single trait and its parameterization.

    Trait objects are collected in a serial integer-keyed dict,
    one per trait that is parameterized for the corresponding Species,
    which inheres to the Species' GenomicArchitecture object as the attribute
    `GenomicArchitecture.traits`, from which individual Traits can be indexed
    using their index-number keys (e.g. `GenomicArchitecture.traits[<idx>]`).

    Attributes
    ----------

        NOTE: For more detail, see the documentation for the parameters
              that correspond to many of the following attributes.

        alpha:
            A 1d numpy array containing the effect sizes of all loci subtending
            this trait. The value at index i represents the effect size of the
            locus at index i in the 'loci' attribute of this Trait.

        alpha_distr_mu:
            Mean of the normal (Gaussian) distribution
            of effect sizes for this trait's loci.

        alpha_distr_sigma:
            Standard deviation of the normal (Gaussian) distribution
            of effect sizes for this trait's loci.

        gamma:
            Curvature of the fitness function.

        idx:
            Index number of the trait (i.e. its key within the traits dict)

        loc_idx:
            A 1d numpy array containing the index positions within an
            Individual's non-neutral genotypes object
            (attribute 'g' of an Individual) that
            correspond to the loci subtending this trait.
            (This is designed such that an Individual's genotype
            for the locus number stored at Trait.loci[i]
            can be accessed by calling Individual.g[Trait.loc_idx[i], :].)
            This method of tracking allows Individuals' full genotypes to be
            stored in the tskit tables while they carry their non-neutral
            genotypes with them (a computational optimization).

        loci:
            A numerically sorted 1d numpy array containing
            the locus numbers of all loci subtending this trait.

        lyr_num:
            Index number of the Landscape layer to which this trait corresponds
            (i.e. of the environmental variable that serves as the selective
            force on this trait)

        max_alpha_mag:
            Maximum absolute value that can be drawn for a locus’
            effect size (i.e. alpha). Effect sizes are clipped
            to the closed interval [-max_alpha_mag, max_alpha_mag]
            when the model is parameterized.

        mu:
            Mutation rate for this trait (expressed in mutations per locus per
            timestep)

        name:
            The string name of the trait

        n_loci:
            Number of loci subtending this trait

        phi:
            Polygenic selection coefficient (i.e the selection coefficient
            acting on the phenotypes, rather than the genotypes, of this Trait)

        univ_adv:
            Whether or not this trait will be treated as universally
            advantageous. When False (the default behavior), individuals will
            be fitter when their phenotypic values for this trait more closely
            align with the local environmental value of the landscape layer
            corresponding to this trait (i.e. selection will be spatially
            varying). When False, phenotypes closer to 1 will be more
            advantageous globally on the Landscape.

    """
    def __init__(self, idx, name, phi, n_loci, mu, layer, alpha_distr_mu,
                 alpha_distr_sigma, max_alpha_mag, gamma, univ_adv):
        self.idx = idx
        self.name = name
        self.phi = phi
        self.n_loci = n_loci
        if mu is None:
            mu = 0
        self.mu = mu
        self.lyr_num = layer
        self.alpha_distr_mu = alpha_distr_mu
        self.alpha_distr_sigma = alpha_distr_sigma
        self.max_alpha_mag = max_alpha_mag
        self.gamma = gamma
        self.univ_adv = univ_adv

        self.loci = np.int64([])
        self.loc_idx = np.int64([])
        self.alpha = np.array([])

    def _get_phi(self, spp):
        if type(self.phi) in (float, int):
            phi = np.array([self.phi]*len(spp))
        else:
            phi = self.phi[spp.cells[:, 1], spp.cells[:, 0]]
        return(phi)

    def _set_loci(self, loci):
        # set the loci
        self.loci = np.hstack((self.loci, np.array([*loci])))
        self.loci.sort()
        # set the number of loci
        self.n_loci = self.loci.size

    def _set_loc_idx(self, nonneut_loci):
        # set an index of the loci's indices in the nonneut_loci object,
        # to use when subsetting for the trait's genotypes
        self.loc_idx = np.array([np.where(
                                nonneut_loci == n)[0][0] for n in self.loci])

    def _add_locus(self, locus, alpha, idx):
        # get insertion index for the locus 
        insert_pt = bisect.bisect_left(self.loci, locus)
        # insert in the loci
        self.loci = np.hstack((self.loci[:insert_pt],
                               locus,
                               self.loci[insert_pt:]))

        # insert the effect size for the locus
        self.alpha= np.hstack((self.alpha[:insert_pt],
                               alpha,
                               self.alpha[insert_pt:]))

        # insert in the loc_idx
        # NOTE: add 1 to all superseding locus indexes, to account for
        # locus newly inserted into genotype arrays
        self.loc_idx = np.hstack((self.loc_idx[:insert_pt],
                                  idx,
                                  self.loc_idx[insert_pt:] + 1))

        # increment the number of loci
        self.n_loci += 1


class GenomicArchitecture:
    """
    Representation of the genomic architecture of a Species.

    Inheres to its Species as the Species.gen_arch attribute.

    Attributes
    ----------

        NOTE: For more detail, see the documentation for the parameters
              that correspond to many of the following attributes.

        delet_alpha_distr_scale:
            The scale parameter of the Gamma distribution from which
            the effect sizes (i.e. alphas) of deleterious loci are drawn

        delet_alpha_distr_shape:
            The shape parameter of the Gamma distribution from which
            the effect sizes (i.e. alphas) of deleterious loci are drawn

        delet_loci:
            A 1d numpy array that tracks all deleterious loci

        delet_loc_idx:
            A 1d numpy array containing the index positions within an
            Individual's non-neutral genotypes object
            (attribute 'g' of an Individual) that
            correspond to the deleterious loci.
            (This is designed such that an Individual's genotype for
            the deleterious locus number stored at
            GenomicArchitecture.delet_loci[i] can be accessed by calling
            Individual.g[GenomicArchitecture.delet_loc_idx[i], :].)
            This method of tracking allows Individuals' full genotypes to be
            stored in the tskit tables while they carry their non-neutral
            genotypes with them (a computational optimization).

        delet_loci_s:
            A 1d numpy array containing the selection strengths
            of each of the deleterious loci. (Selection strengths are
            listed in an order corresponding to the order of
            GenomicArchitecture.delet_loci, such that the selection strength of
            the deleterious locus referenced at
            GenomicArchitecture.delet_loci[i] is stored at
            GenomicArchitetcture.delet_loci_s[i].)
            
        dom:
            Dominance values for all loci (stored as an L-length,
            1d numpy array in which 0 indicates a codominant locus
            and 1 indicates a locus for which the '1' allele is dominant)

        L:
            Length of the simulated genome

        mu_delet:
            Genome-wide deleterious mutation rate (expressed in mutations
            per locus per timestep). Deleterious mutations do not influence
            trait phenotypes, but instead are univerally
            deleterious (to an extent determined by their effect sizes).

        mu_neut:
            Genome-wide neutral mutation rate (expressed in mutations
            per locus per timestep).
        
        neut_loci:
            A 1d numpy array that tracks all loci that do not influence the
            phenotypes of any traits.

        nonneut_loci:
            A 1d numpy array that tracks all loci that either influence the
            phenotype of at least one trait or are deleterious.

        p:
            Starting allele frequencies for all loci (stored as an L-length,
            1d numpy array)

        pleiotropy:
            A bool indicating whether or not to allow pleiotropy (i.e. whether
            or not to allow the same locus to subtend multiple traits)

        recombinations:
            The Recombinations object, which contains the parameters and data
            necessary for simulation of recombination. Not intended for public
            use at this time.

        sex: 
            A bool indicating whether or not Individuals of this Species
            will be assigned a sex (0 for female, 1 for male)

        traits:
            A dict containing all of a Species' Trait objects, keyed by serial
            integers.
         
        x:
            Ploidy
            NOTE: only diploidy is currently implemented, but this attribute
                  serves as a placeholder for use in possible
                  future generalization to arbitrary ploidy.)

    """

    def __init__(self, dom, g_params, land, recomb_rates=None,
                 recomb_positions=None):
        # ploidy (NOTE: for now will be 2 by default; later could consider
        # enabling polyploidy)
        self.x = 2
        # total length (i.e. number of markers)
        self.L = g_params.L
        # placeholder for the starting allele freqs for all loci
        self.p = None
        # True/False regarding whether to allow a locus to affect the
        # phenotype of more than one trait; defaults to False
        self.pleiotropy = g_params.pleiotropy
        # array of dominance values for all loci
        self.dom = dom
        # set the _use_dom attribute based on whether any loci have a 1 for
        # their dominance value
        self._use_dom = np.any(self.dom)
        # whether or not to use sexes for this species
        self.sex = g_params.sex

        # genome-wide neutral mutation rate
        self.mu_neut = g_params.mu_neut
        # array to keep track of all loci that don't influence the
        # phenotypes of any trait; defaults to all loci, then will be updated
        self.neut_loci = np.array(range(self.L))
        # array to keep track of all loci that influence the phenotype of at
        # least one trait; after burn-in, will be updated as needed
        self.nonneut_loci = np.array([])

        # genome-wide deleterious mutation rate
        self.mu_delet = g_params.mu_delet
        self.delet_alpha_distr_shape = g_params.delet_alpha_distr_shape
        self.delet_alpha_distr_scale = g_params.delet_alpha_distr_scale


        # arrays to track deleterious loci, their genotype indices, and their
        # strengths of selection
        self.delet_loci = np.int64([])
        self.delet_loc_idx = np.int64([])
        self.delet_loci_s = np.array([])

        # add a dict of Trait objects, if necessary
        self.traits = None
        if 'traits' in [*g_params]:
            self.traits = _make_traits(g_params.traits, land)

        # set self._mu_tot, the total per-site, per-generation mutation rate
        mus = [mu for mu in (self.mu_neut, self.mu_delet) if mu is not None]
        if self.traits is not None:
            mus = mus + [trt.mu for trt in self.traits.values()]
        self._mu_tot = sum(mus)
        # save the nonneutral mutation rate
        self._mu_nonneut = self._mu_tot - self.mu_neut

        # set a placeholder for the species' mutable loci
        # (to be filled after burn-in)
        self._mutables = None

        # attribute that will be replaced with the estimated total number of
        # mutations per iteration, once it's estimated at the end of the first
        # burn-in
        self._mut_fns = self._make_mut_fns_dict()
        # set ._planned_muts to None, for now (this is not yet implemented,
        # but thinking about it
        self._planned_muts = None

        # The recombination-paths object will be assigned here; used to
        # speed up large quantities of binomial draws needed for recombination
        self.recombinations = Recombinations(self.L, recomb_positions,
                                             g_params.n_recomb_sims,
                                             g_params.r_distr_alpha,
                                             g_params.r_distr_beta,
                                             recomb_rates,
                                             g_params.jitter_breakpoints)

    # method to make a _mut_fns dict, containing a function
    # for each type of mutation for this species
    def _make_mut_fns_dict(self):
        mut_fns = {}
        if self.mu_neut > 0:
            def neut_fn(spp, offspring):
                return(mutation._do_neutral_mutation(spp, offspring))
            mut_fns.update({'neut': neut_fn})
        if self.mu_delet > 0:
            def delet_fn(spp, offspring):
                return(mutation._do_deleterious_mutation(spp, offspring))
            mut_fns.update({'delet': delet_fn})
        if self.traits is not None:
            for trait_num in self.traits:
                if self.traits[trait_num].mu > 0:
                    def trait_fn(spp, offspring, trait_num=trait_num):
                        return(mutation._do_trait_mutation(spp, offspring,
                                                           [trait_num]))
                    mut_fns.update({'t%i' % trait_num: trait_fn})
        return mut_fns

    # method to draw mutation types for any number of mutations chosen
    # to occur in a given timestep
    def _draw_mut_types(self, num):
        type_dict = {'neut': self.mu_neut,
                     'delet': self.mu_delet}
        if self.traits is not None:
            trait_dict = {'t%i' % (k): v.mu for k, v in self.traits.items()}
            type_dict.update(trait_dict)
        types = []
        probs = []
        for k, v in type_dict.items():
            types.append(k)
            probs.append(v)
        probs = [p/sum(probs) for p in probs]
        choices = r.choice(types, p=probs, size=num, replace=True)
        return(choices)

    # method for drawing an effect size for one or many loci
    def _draw_trait_alpha(self, trait_num, n=1):
        mu = self.traits[trait_num].alpha_distr_mu
        sigma = self.traits[trait_num].alpha_distr_sigma
        max_alpha_mag = self.traits[trait_num].max_alpha_mag
        if max_alpha_mag is not None:
            min_alpha = -1 * max_alpha_mag
            max_alpha = max_alpha_mag
        else:
            min_alpha = max_alpha = max_alpha_mag
        # use mu value as the fixed effect size, if sigma is 0
        if sigma == 0:
            alpha = mu * np.array([1 - (i % 2)*2 for i in range(n)])
        else:
            alpha = r.normal(self.traits[trait_num].alpha_distr_mu,
                             self.traits[trait_num].alpha_distr_sigma, n)
            if max_alpha_mag is not None:
                alpha = np.clip(alpha, min_alpha, max_alpha)
        # otherwise use mu and sigma to draw effects from
        # (because effects will be added to 0)
        if self.traits[trait_num].n_loci == 1:
            alpha = np.abs(alpha)
        return(alpha)

    # method for drawing new deleterious mutational fitness effects
    def _draw_delet_s(self):
        s = r.gamma(self.delet_alpha_distr_shape, self.delet_alpha_distr_scale)
        s = min(s, 1)
        return(s)

    # method for assigning loci to traits
    def _set_trait_loci(self, trait_num, mutational=False,
                        loci=None, alpha=None):
        # if this is not the result of a point mutation, but instead
        # either an initial setup or manually introduced, then grab the
        # number of loci to be assigned
        if not mutational:
            n = self.traits[trait_num].n_loci
            assert n <= self.L, ("The number of loci parameterized for "
                                 "trait number %i ('n_loci') is greater "
                                 "than the length of the genome!")
        # otherwise, assign a single locus
        else:
            n = 1
        if loci is not None:
            if not np.iterable(loci):
                loci = [loci]
            assert len(set([*loci])) == len(loci), ("Some of the trait "
                                                    "loci provided appear "
                                                    "to be repeated.")
        # else, draw loci randomly, either allowing pleiotropy or not
        elif not self.pleiotropy:
            loci = set(r.choice(self.neut_loci, size=n, replace=False))
        elif self.pleiotropy:
            loci = set(r.choice(range(self.L), size=n, replace=False))
        # update the trait's loci
        self.traits[trait_num]._set_loci(loci)
        # add these loci to self.non-neutral and remove from
        # self.neut_loci, to keep track of all loci underlying any
        # traits (for purpose of avoiding pleiotropy)
        self.nonneut_loci = np.array(sorted([*self.nonneut_loci] + [*loci]))
        self.neut_loci = np.array(sorted([*set(self.neut_loci).difference(
                                                    set(self.nonneut_loci))]))
        # if the effect size(s) is/are provided, use those
        if alpha is not None:
            if not np.iterable(alpha):
                alpha = np.array([alpha])
            effects = np.array([*alpha])
        # else, draw effects from a Gaussian dist with mean 0 and sigma
        # provided by trait params (as per e.g. Yeaman and Whitlock 2011)
        else:
            effects = self._draw_trait_alpha(trait_num, n)
        # if this is drawing the effect of a monogenic trait's locus (i.e. if
        # it's drawing only one value, but it's not mutational), then coerce
        # the effects to [0.5]
        if not mutational and n == 1:
            effects = np.array([0.5])
        # check that loci and effects are of equal length
        assert len(loci) == len(effects), ('Lengths of the two arrays '
                                           'containing the new trait loci and '
                                           'their effects are not equal.')
        # then add the effects to the trait's alpha array
        self.traits[trait_num].alpha = np.hstack((self.traits[trait_num].alpha,
                                                  effects))

    # add a nonneutral locus to the genomic architecture
    # NOTE: either trait_nums or delet_s must be non-None,
    # and trait_nums must be iterable (if not None)
    def _add_nonneut_locus(self, locus, trait_nums=None, delet_s=None):
        # remove from the neut_loci array
        self.neut_loci = np.delete(self.neut_loci,
                                   np.where(self.neut_loci == locus))
        # add the locus to the nonneut_loci array
        idx = bisect.bisect_left(self.nonneut_loci, locus)
        self.nonneut_loci = np.hstack((self.nonneut_loci[:idx],
                                       locus,
                                       self.nonneut_loci[idx:]))

        # add the locus to either trait loci, if necessary
        if trait_nums is not None and delet_s is None:
            for n in trait_nums:
                alpha = self._draw_trait_alpha(n)[0]
                self.traits[n]._add_locus(locus, alpha, idx)
        # or else add to the deleterious loci
        elif delet_s is not None and trait_nums is None:
            del_idx = bisect.bisect_left(self.delet_loci, locus)
            # add the locus, its genome-index (for subsetting individuals'
            # genomes when calculating fitness), and its strength of selection
            # to the deleterious locus trackers
            self.delet_loci = np.hstack((self.delet_loci[:del_idx],
                                         locus,
                                         self.delet_loci[del_idx:]))
            self.delet_loc_idx = np.hstack((self.delet_loc_idx[:del_idx],
                                         idx,
                                         self.delet_loc_idx[del_idx:]))
            self.delet_loci_s = np.hstack((self.delet_loci_s[:del_idx],
                                         delet_s,
                                         self.delet_loci_s[del_idx:]))

        #TODO REMOVE NEXT 2 LINES AFTER TESTING
        else:
            assert True == False, "BOTH TRAITS_NUMS AND DELET_S CANT BE NONE!"

        return idx


    # method for plotting all allele frequencies for the species
    def _plot_allele_frequencies(self, spp):
        speciome = np.stack([ind.g for ind in spp.values()])
        freqs = speciome.sum(axis=2).sum(axis=0) / (2*speciome.shape[0])
        plt.plot(range(self.L), self.p, ':r', label='start freq.')
        plt.plot(range(self.L), freqs, '-b', label='curr. freq.')
        plt.xlabel('locus')
        plt.ylabel('frequency')
        plt.legend()
        plt.show()

    # method for pickling a genomic architecture
    def _write_pickle(self, filename):
        import cPickle
        with open(filename, 'wb') as f:
            cPickle.dump(self, f)


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

# generate allele_freqs
def _draw_allele_freqs(l):
    return(r.beta(1, 1, l))


def _make_traits(traits_params, land):
    params_copy = {**traits_params}
    # and set each Layer number using Layer names
    for k, v in params_copy.items():
        # the first time this is run during a model with random
        # communities at each iteration, the layer identified in
        # the traits_params will be a string indicating the layer's name
        if isinstance(v.layer, str):
            lyr_num = [num for num, lyr in land.items(
                                                ) if lyr.name == v.layer]
        # the second and later times it's run during such a model (see
        # previous comment), it will already have been swapped out for an int
        # indicating the layer's index num
        elif isinstance(v.layer, int):
            lyr_num = [num for num, lyr in land.items(
                                                ) if lyr.idx == v.layer]
        assert len(lyr_num) == 1, ("Expected to find a single Layer with "
                                   "the Layer name indicated for Trait %s, "
                                   "but instead found "
                                   "%i.") % (k, len(lyr_num))
        v['layer'] = lyr_num[0]
    # then for each of i traits, unpack the ith components of the remaining
    # params to create the trait dict
    traits = {n: Trait(n, k_v[0], **k_v[1]) for n, k_v in enumerate(
                                                        params_copy.items())}
    # for all monogenic traits, if the trait doesn't already have a 0 mutation
    # rate then coerce it to 0
    # NOTE: this is the because there is a good reason to use 0 as the 
    # baseline phenotype for monogenic traits but 0.5 as the baseline
    # for multilocus traits (to preserve symmetry of ability for individuals
    # to have phenotypes beyond 0 and 1 for a multilocus trait and thus
    # experience stabilizing selection within their optimal habitat), but
    # this means that there would be a huge problem if a monogenic trait
    # underwent an adaptive mutation and became polyenic bceause all of the
    # individuals' phenotypes would suddenly have to be recalculated and would
    # suddenly completely change
    for n, trt in traits.items():
        if trt.n_loci == 1:
            if trt.mu != 0:
                warnings.warn(("Coercing Trait %i ('%s') to a "
                               "0 mutation rate because it is monogenic."))
                trt.mu = 0
    return(traits)


# build the genomic architecture
def _make_genomic_architecture(spp_params, land):
    # get the genome parameters
    g_params = spp_params.gen_arch
    # get the custom genomic-architecture file, if provided
    gen_arch_file = None
    if 'gen_arch_file' in g_params.keys():
        if g_params.gen_arch_file is not None:
            gen_arch_file = pd.read_csv(g_params.gen_arch_file)
            # ensure that trait and alpha columns are strings to start (because
            # in many cases alpha would likely read in as a float, if each row
            # has at most one alpha value because there's no pleiotropy, but in
            # case of pleiotropy I need to be able to use the str.split method
            # on each row's values further down)
            gen_arch_file['trait'] = [str(v) for v in gen_arch_file['trait']]
            gen_arch_file['alpha'] = [str(v) for v in gen_arch_file['alpha']]
            assert (np.all(
                (gen_arch_file['dom'] == 0) + gen_arch_file['dom'] == 1)), (
                "The 'dom' column of the custom genomic architecture file "
                "must contain only 0s and 1s (where 0 indicates codominance "
                "of the 0 and 1 alleles, 1 indicates that the 1 allele "
                "is dominant).")
            # check that each trait has in that file the number of loci
            # indicated by n_loci in that trait's params dict
            if 'traits' in [*g_params]:
                all_traits = [trt.strip() for row in [val.split(
                    ',') for val in gen_arch_file['trait']] for trt in row]
                for trt_name, trt in g_params.traits.items():
                    n_loci_in_file = sum(
                                    [trt == trt_name for trt in all_traits])
                    assert_msg = ("The number of times a Trait is appears "
                                  "in the custom genomic architecture file "
                                  "must be equivalent to the number of loci "
                                  "subtending that Trait as indicated by the "
                                  "'n_loci' key in its section of the "
                                  "parameters file.")
                    assert n_loci_in_file == trt.n_loci, assert_msg

    # get the sex parameter and add it as an item in g_params
    g_params['sex'] = spp_params.mating.sex

    # get the custom recomb_rates and recomb_positions from the
    # custom gen-arch file, if provided
    # NOTE: add 0.5 to each locus' position, because each recombination rate
    #       stipulated in gen_arch file is construed as the recombination rate
    #       at the midway point between that locus and the subsequent one
    recomb_rates = None
    recomb_positions = None
    if gen_arch_file is not None:
        recomb_rates = gen_arch_file['r'].values[:-1]
        recomb_positions = gen_arch_file['locus'].values[:-1] + 0.5

    # set locus-wise dominance values for the 1-alleles, using the 'dom' value
    # in the gen_arch params, unless a gen_arch_file was provided
    if gen_arch_file is None:
        # create an L-length array of boolean integers (where 0 = codominance,
        # 1 = dominance)
        dom = np.array([int(g_params.dom)] * g_params.L)
    else:
        # get the 'dom' column from the gen_arch_file
        dom = gen_arch_file['dom'].values

    # now make the gen_arch object
    gen_arch = GenomicArchitecture(dom, g_params, land,
                                   recomb_rates, recomb_positions)

    # set the loci and effect sizes for each trait, using the custom gen-arch
    # file, if provided, and if the species has traits
    if gen_arch_file is not None and gen_arch.traits is not None:
        # convert the trait names in the 'trait' column of the file into
        # lists of their trait numbers (i.e. their keys in the gen_arch
        # traits dict)
        trt_names_nums = {
            trt.name: num for num, trt in gen_arch.traits.items()}
        gen_arch_file['trait'] = [[trt_names_nums[
            val] for val in [x.strip() for x in row.split(
            ',')] if val in trt_names_nums] for row in gen_arch_file['trait']]
        # turn the values in the 'alpha' column into lists of
        # values, by splitting on commas
        # (this will allow people to assign a single locus
        # to more than one trait, i.e. to model pleiotropy)
        gen_arch_file['alpha'] = [
            [*map(float, row.split(','))] for row in gen_arch_file['alpha']]
        # get the loci and effect sizes for each trait
        loci = {}
        alphas = {}
        for trt_num in gen_arch.traits.keys():
            loci[trt_num] = np.array([*gen_arch_file['locus']])[
                [trt_num in row for row in gen_arch_file['trait']]]
            alphas[trt_num] = np.concatenate([np.array(row['alpha'])[
                [n == trt_num for n in row[
                    'trait']]] for i, row in gen_arch_file.iterrows()])
        # check that we got the same length of loci and effect sizes for
        # a trait, for all traits
        for trt_num in loci.keys():
            assert_msg = ("Expected to receive the same number of loci and "
                          "alphas (i.e. effect sizes) for trait number %i, "
                          "but instead got %i loci and %i "
                          "alphas.") % (trt_num, len(loci[trt_num]),
                                        len(alphas[trt_num]))

            assert len(loci[trt_num]) == len(alphas[trt_num]), assert_msg

        # add the loci and effect sizes for each of the traits to the
        # Trait object in the GenomicArchitecture
        for trait_num in gen_arch.traits.keys():
            gen_arch._set_trait_loci(trait_num, mutational=False,
                                     loci=loci[trait_num],
                                     alpha=alphas[trait_num])
    # or else randomly set the loci and effect sizes for each trait
    else:
        if gen_arch.traits is not None:
            for trait_num in gen_arch.traits.keys():
                gen_arch._set_trait_loci(trait_num, mutational=False)

    # now that all nonneutral loci have been selected, 
    # set each trait's loc_idx (which maps a trait's numeric loci to
    # their index positions in individuals' genomes)
    if gen_arch.traits is not None:
        for trt in gen_arch.traits.values():
            trt._set_loc_idx(gen_arch.nonneut_loci)

    assert_msg = ("The union of the gen_arch.neut_loci and "
                  "gen_arch.nonneut_loci sets does not contain all loci "
                  "indicated by gen_arch.L")
    assert len(set(range(gen_arch.L)).difference(
        set(gen_arch.neut_loci).union(
        set(gen_arch.nonneut_loci)))) == 0, assert_msg

    # draw locus-wise 1-allele frequencies, unless provided in
    # custom gen-arch file
    if gen_arch_file is None:
        # set according to a non-None value
        if g_params.start_p_fixed is not None:
            if isinstance(g_params.start_p_fixed, bool):
                # if True, make all 0.5 (default starting allele freq)
                if g_params.start_p_fixed:
                    gen_arch.p = np.array([0.5]*g_params.L)
                # if False, draw randomly
                else:
                    gen_arch.p = _draw_allele_freqs(g_params.L)
            # if not a bool then should be a float, so assert, then set to that val
            else:
                assert 0 <= g_params.start_p_fixed <= 1, ("If a starting allele "
                                                          "frequency value is "
                                                          "provided then it must "
                                                          "be between 0 and 1.")
                gen_arch.p = np.array([g_params.start_p_fixed]*g_params.L)
        # if the value is None, then draw randomly
        else:
            gen_arch.p = _draw_allele_freqs(g_params.L)
        # set the neutral loci to 0, if need be
        if g_params.start_neut_zero:
            gen_arch.p[gen_arch.neut_loci] = 0
    # use starting allele freq values from the gen_arch_file instead, if given
    else:
        gen_arch.p = gen_arch_file['p'].values

    # set the recombination events
    use_subsetters = (len(gen_arch.nonneut_loci) > 0 or
                      gen_arch._mu_nonneut > 0)
    gen_arch.recombinations._set_events(use_subsetters, gen_arch.nonneut_loci)

    return gen_arch


# reset genomes after burn-in
def _check_mutation_rates(gen_arch, est_tot_muts, burn_T, T):
    # if there do not appear to be enough space in the simulated genome for
    # the expected number of mutations then warn the user of that
    if est_tot_muts > 0.75 * len(
        [loc for loc in range(
                    gen_arch.L) if loc not in gen_arch.nonneut_loci]):
        raise MutationRateError(("This species has been "
                                 "parameterized with too few neutral "
                                 "loci to accommodate the expected "
                                 "number of mutations. (Geonomics only "
                                 "uses an infinite sites model.) "
                                 "Please tweak some combination of "
                                 "the genome length, model run time, "
                                 "or mutation rates."))


    # skip this step and force the neutral mutation rate to 0, if there are no
    # neutral loci in the genome as it was configured
    if len(gen_arch.neut_loci) == 0:
        raise MutationRateError(("This species has been parameterized "
                                 "with non-zero mutation rates but "
                                 "without any neutral loci, leaving no target "
                                 "for mutations. Please tweak the genome "
                                 "length and/or mutation rates."))
        gen_arch.mu_neut = 0
        gen_arch.mu_delet = 0
        for trt in gen_arch.traits.values():
            trt.mu = 0
    # or just outright skip this step if the parameterization included only
    # mutation rates of 0
    elif gen_arch._mu_tot == 0:
        pass
    # otherwise, set the mutable loci
    else:
        mutables = [*set(range(gen_arch.L)).difference(
                                            set(gen_arch.nonneut_loci))]
        r.shuffle(mutables)
        gen_arch._mutables = [*mutables]
    return


# function to generate mutations, after burn-in,
# and to assign them to a species' TableCollection's  current nodes,
# to produce the starting 1-allele frequencies parameterized for the species
def _make_starting_mutations(spp, tables):
    # get the starting frequencies for each site
    start_freqs = spp.gen_arch.p

    # get a set of all homologues, as tuples of (individual id, homologue idx)
    homologues = [*zip(np.repeat([*spp], 2),
                         [*range(spp.gen_arch.x)] * len(spp))]

    # make mutations for each site
    for site, freq in enumerate(start_freqs):
        # generate the number of mutations for this locus
        n_mutations = int(round(2 * len(spp) * freq, 0))
        # make sure we don't mutate either all or none of the population's
        # homologues, unless called for
        if n_mutations == len(spp) * 2 and freq < 1:
            n_mutations -= 1
        if n_mutations == 0 and freq > 0:
            n_mutations = 1
        #randomly choose and mutate n_mutations homologues from the population 
        np.random.shuffle(homologues)
        homologues_to_mutate = homologues[:n_mutations]
        for ind, homol in homologues_to_mutate:
            # create a mutation in the individual's genome, if this is a
            # non-neutral locus
            if site in spp.gen_arch.nonneut_loci:
                spp[ind].g[np.where(spp.gen_arch.nonneut_loci == site),
                           homol] = 1
            # get the homologue's nodes-table id,
            # then add a row to the mutations table
            node_id = spp[ind]._nodes_tab_ids[homol]
            tables.mutations.add_row(site, node=node_id, parent=-1,
                                     derived_state='1')

    # and then reset the individuals' phenotypes, if needed
    if spp.gen_arch.traits is not None:
        [ind._set_z(spp.gen_arch) for ind in spp.values()]

    return


# method for loading a pickled genomic architecture
def read_pickled_genomic_architecture(filename):
    import cPickle
    with open(filename, 'rb') as f:
        gen_arch = cPickle.load(f)
    return gen_arch


##########################################################################
# FUNCTIONS FOR WORKING WITH THE SPATIAL PEDGREE SAVED IN tskit STRUCTURES
##########################################################################

def _get_lineage_dicts(spp, nodes, loci, t_curr, drop_before_sim=True,
                       time_before_present=True, max_time_ago=None,
                       min_time_ago=None):
    '''
    For a sample of nodes and a sample of loci, get the
    loci-sequentially ordered dict of lineage dicts
    (containing each node in a child node's lineage as the
    key and its birth-time and birth-location as a length-2 list of values

    Thus, the returned value is of the structure:
        {locus_id_1: {node_id_1: {curr_node_id: (birth_time,
                                             array([birth_loc_par_1,
                                                    birth_loc_par_2])),
                              prev_node_id: (birth_time,
                                             array([birth_loc_par_1,
                                                    birth_loc_par_2])),
                              .
                              .
                              .
                              oldest_node_id: (birth_time,
                                             array([birth_loc_par_1,
                                                    birth_loc_par_2]))},
                    .
                    .
                    .
                    node_id_last: {curr_node_id: ...
                                    ...
                                    }},
        .
        .
        .
        locus_id_last: {node_id_1: {curr_node_id: ...
                                    ...
                                    }}
        }
    '''

    # make sure the TableCollection is sorted, then get the TreeSequence
    try:
        ts = spp._tc.tree_sequence()
    except Exception:
        raise Exception(("The species' TableCollection must be sorted and "
                         "simplified before this method is called."))
    # get the tree numbers corresponding to each locus
    treenums = _get_treenums(ts, loci)
    # get the trees
    trees = ts.aslist()
    # create the output data structure
    lineage_dicts = {}
    # get the lineage_dict for each locus
    for loc, treenum in zip(loci, treenums):
        # store the lineage dicts for the current locus
        lineage_dicts_curr_loc = _get_lineage_dicts_one_tree(spp._tc,
                                    trees[treenum], nodes, t_curr,
                                    drop_before_sim=drop_before_sim,
                                    time_before_present=time_before_present,
                                    max_time_ago=max_time_ago,
                                    min_time_ago=min_time_ago)
        lineage_dicts[loc] = lineage_dicts_curr_loc
    return lineage_dicts


def _get_lineage_dicts_one_tree(tc, tree, nodes, t_curr, drop_before_sim=True,
                                time_before_present=True, max_time_ago=None,
                                min_time_ago=None):
    '''
    For a sample of nodes, and for a given tree,
    get dicts of each node's lineage nodes with their birth locs and birth
    times
    '''

    #create an output master dict
    lineage_dicts = {}
    # get each sample node's lineage dict
    for node in nodes:
        lineage = _get_lineage(tree, [node])
        lineage_dict =dict(zip(lineage, _get_lineage_times_and_locs(
                                    tc, lineage, t_curr,
                                    drop_before_sim=drop_before_sim,
                                    time_before_present=time_before_present)))

         # filter for time, if requested
        if max_time_ago is not None or min_time_ago is not None:
            # if only one time limit was set, set the other correctly, based on
            # whether or not time is expressed as time before present
            if max_time_ago is None:
                max_time_ago = np.inf
            if min_time_ago is None:
                min_time_ago = -np.inf
            # filter by time
            lineage_dict = {node: val for node, val in lineage_dict.items() if (
                            min_time_ago <= val[0] <= max_time_ago)}

        lineage_dicts[node] = lineage_dict

    return lineage_dicts


# get birth-locations and times of nodes along a pedigree
def _get_lineage_times_and_locs(tc, lineage, t_curr, drop_before_sim=True,
                                time_before_present=True):
    locs = []
    times = []
    for node in lineage:
        time = tc.nodes[node].time
        # NOTE: optionally drop all nodes from before the simulation
        # (i.e. nodes which were added to TableCollection by
        # backward-time simulation with ms); THIS IS THE DEFAULT
        # (or include all, if drop_before_sim is False)
        if (drop_before_sim and time < 0) or not drop_before_sim:
            times.append(time)
            individ = tc.nodes[node].individual
            loc = tc.individuals[individ].location
            locs.append(loc)
    # express times as time before present, if requested
    if time_before_present:
        # NOTE: need to make t_curr negative, to reflect the fact that time is
        # stored as positive integeres in Geonomics data structures but as
        # negative integers in tskit TableCollection (because they require time
        # expressed as time before present, so using positive integers would
        # not allow models to be run indefinitely, for exploratory/fun purposes
        times = [t - -t_curr for t in times]
    return(zip(times, locs))


def _get_lineage(tree, nodes):
    '''
    Recursive algorithm for getting all parents of a node
    '''
    parent = tree.parent(nodes[-1])
    if parent == tskit.NULL:
        return(nodes)
    else:
        nodes.append(parent)
        return(_get_lineage(tree, nodes))


# get a dict of the interval (i.e. tree) number
# keyed to each locus in a list of loci
def _get_treenums(ts, loci, as_dict=False):
    interval_right_ends = [t.get_interval()[1] for t in ts.trees()]
    treenums = [bisect.bisect_left(interval_right_ends, loc) for loc in loci]
    if as_dict:
        treenums = {loc: treenum for loc, treenum in zip(loci, treenums)}
    return(treenums)


# calculate a stat for the lineage corresponding to the provided lineage_dict
def _calc_lineage_stat(lin_dict, stat):
    assert stat in ['dir', 'dist', 'time', 'speed'], ("The only valid "
                                                       "statistics are: "
                                                       "direction ('dir'), "
                                                       "distance ('dist'), "
                                                       "time ('time'), "
                                                       "and speed ('speed').")
    fn_dict = {'dir': _calc_lineage_direction,
               'dist': _calc_lineage_distance,
               'time': _calc_lineage_time,
               'speed': _calc_lineage_speed
              }
    val = fn_dict[stat](lin_dict)
    return val


# calculate the angular direction of gene flow along a locus' lineage
def _calc_lineage_direction(lin_dict):
    # if the lineage has 1 or less node, don't calculate anything
    if len(lin_dict.values()) < 2:
        return
    else:
        # get x and y distances between beginning and ending points
        # (begins at the bottom of the lineage dict, furthest back in time)
        beg_loc = [*lin_dict.values()][-1][1]
        # (ends at the top of the lineage dict, in the current time step)
        end_loc = [*lin_dict.values()][0][1]
        # get the difference in the x and y axes between the lineage's
        # beginning location and its ending (i.e. current) location
        # NOTE: using range(2) will grab only the first location values,
        # which is perfect, because models with traits also save phenotypic
        # and fitness values in the individuals table's 'location' column
        x_diff, y_diff = [end_loc[i] - beg_loc[i] for i in range(2)]
        # get the counterclockwise angle, expressed in degrees
        # from the vector (X,Y) = (1,0),
        # with 0 to 180 in quadrants 1 & 2, 0 to -180 in quadrants 3 & 4
        ang = np.rad2deg(np.arctan2(y_diff, x_diff))
        # convert to all positive values, 0 - 360
        if ang < 0:
            ang += 360
        #convert to all positive clockwise angles, expressed as 0 at compass north
        # (i.e. angles clockwise from vector (X,Y) = (0,1)
        ang = (-ang + 90) % 360
        return ang


# calculate the geographic distance (in cell widths)
# of gene flow along a locus' lineage
def _calc_lineage_distance(lin_dict):
    if len(lin_dict.values()) < 2:
        return
    else:
        # get x and y distances between beginning and ending points
        beg_loc = [*lin_dict.values()][0][1]
        end_loc = [*lin_dict.values()][-1][1]
        x_diff, y_diff = [end_loc[i] - beg_loc[i] for i in range(2)]
        #Pythagoras
        dist = np.sqrt(x_diff**2 + y_diff**2)
        return dist


# calculate the total time to the simulation's MRCA of a locus' lineage
def _calc_lineage_time(lin_dict):
    if len(lin_dict.values()) < 2:
        return
    else:
        beg_time = [*lin_dict.values()][0][0]
        end_time = [*lin_dict.values()][-1][0]
        time = end_time - beg_time
        return time


# calculate the speed of gene flow in a lineage (in cell widths/time steps)
def _calc_lineage_speed(lin_dict):
    if len(lin_dict.values()) < 2:
        return
    else:
        dist = _calc_lineage_distance(lin_dict)
        time = _calc_lineage_time(lin_dict)
        speed = dist/time
        return speed



#!/usr/bin/python
# main.py

# flake8: noqa

"""
Geonomics
=========

Provides:
1. A function for creating custom-structured, editable parameters files.
2. A function for creating model objects from parameters files.
3. Functions for running models.
4. Functions for model visualization.
5. A variety of parameters, for construction of arbitrarily complex simulation scenarios.

How to use the documentation
----------------------------
There are two levels of documentation available:
- Basic, procedural documentation is available as docstrings in the code,
which are returned when requesting Python's help on a class or function.
- Detailed documentation is available at the
`Geonomics homepage <https://geonomics.readthedocs.io/en/latest/>`_.

All documentation examples assume that `geonomics` has been imported as `gnx`::

    >>> import geonomics as gnx

To get help, call Python's `help` function on a component of Geonomics that you
are trying to use, such as::

    >>> help(gnx.make_parameters_file)

Or use the question mark, if you are working in the
`IPython <https://ipython.org>`_ shell or in the
`Jupyter <https://jupyter.org/>`_ Notebook::

    >>> geonomics.make_parameters_file?
"""
#geonomics imports
from geonomics.sim.model import Model
from geonomics.sim.params import (_read_params_file, _make_params_file,
                                  ParametersDict)
from geonomics.structs.landscape import _make_landscape
from geonomics.structs.genome import _make_genomic_architecture
from geonomics.structs.individual import _make_individual
from geonomics.structs.species import _make_species
from geonomics.structs.community import _make_community
from geonomics.structs import landscape, genome, individual, species, community
from geonomics import demos

#other imports
import re
import os, sys, traceback
from collections import Counter as C
import numpy as np
import pandas as pd


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

#wrapper around params.make_parameters_file
def make_parameters_file(filepath=None, layers=1, species=1, data=False,
        stats=False):
    """
    Create a new parameters file.

    Write to disk a new, template parameters file. The file will contain the
    numbers and types of sections indicated by the parameters fed to this
    function. The new file can then be used 'out of the box'
    to make a new Model object, but typically it will be
    edited by the user to stipulate the scenario being simulated,
    then used to instantiate a Model.

    Parameters
    ----------
    filepath : str, optional, default: None
        Where to write the resulting parameters file, in /path/to/filename.py
        format. Defaults to None. If None, a file named
        "GNX_params_<datetime>.py" will be written to the working
        directory.

    layers : {int, list of dicts}, optional, default: 1
        Number (and optionally, types) of Layer-parameter sections to include
        in the parameters file that is generated. Defaults to 1. Valid values
        and their associated behaviors are:

        *If an int is passed*:
            Add sections for the stipulated number of Layers, each with default
            settings:
                - parameters for creating Layers of type 'random' (i.e.
                  Layers that will be generated by interpolation from
                  randomly valued random points)
                - no LandscapeChanger parameters

        *If [dict, ..., dict] is passed*:
            Each dict in this list should have the following key-value pairs:
                KEY                    VALUE
                'type':                {'random', 'defined', 'file', 'nlmpy'}

                'change':               bool

            This will add one section of Layer parameters, with the
            contents indicated, for each dict in this list.

    species : {int, list of dicts}, optional, default: 1
        Number (and optionally, types) of Species-parameter sections to
        include in the parameters file that is generated. Defaults to 1. Valid
        values and their associated behaviors are:

        *If an int is passed*:
            Add sections for the stipulated number of Species, each with
            default settings:
                - parameters for movement and dispersal without
                  _ConductanceSurfaces
                - parameters for a GenomicArchitecture with 0 Traits (i.e. with
                  only neutral loci)
                - no _SpeciesChanger parameters

        *If [dict, ..., dict] is passed*:
            Each dict in this list can contain any of the following
            key-value pairs:
                KEY                     VALUE

                'movement':             bool,

                'movement_surface':     bool,

                'dispersal_surface':    bool,

                'genomes':              {bool, 'custom'},

                'n_traits':             int,

                'demographic_change':   int,

                'parameter_change':     bool

            This will add one section of Species parameters, customized
            as indicated, for each dict in the list. (Note that if the
            'genomes' argument is True or 'custom', a section for
            parameterization of the genomic architecture will be added,
            but if it is 'custom' then a template custom genomic architecture
            file will also be created (a CSV file), which can be filled in
            to stipulate the locus-wise values for starting allele frequency,
            recombination rate, dominance, associated traits, and effect
            sizes.)

    data : bool, optional, default: False
        Whether to include a Data-parameter section in the parameters file that
        is generated. Defaults to None. Valid values and their associated
        behaviors are:

        None, False:
            Will not add a section for parameterizing data to be collected.
            No _DataCollector will be created for the Model object made from
            the resulting parameters file, and no data will be collected
            during the model runs.
        True:
            Will add a section that can be used to parameterize which
            data will be collected during the model runs, when, and what
            file formats will be used to write it to disk.
            (This which will be managed by the model's _DataCollector
            object.)

    stats : bool, optional, default: False
        Whether to include a Stats-parameter section in the parameters file that
        is generated. Defaults to None. Valid values and their associated
        behaviors are:

        None, False:
            Will not add a section for parameterizing the statistics to be
            calculated. No _StatsCollector will be created for the Model
            object made from the resulting parameters file, and no
            statistics will be calculated during the model runs.
        True:
            Will add a section that can be used to parameterize which
            statistics will be calculated during the model runs, and when.
            (This will be managed by the model's _StatsCollector object.)

    Returns
    -------
    None
        Returns no output. Resulting parameters file will be written to the
        location and filename indicated (or by default, will be written to a
        file named "GNX_params_<datetime>.py" in the working directory).

    See Also
    --------
    sim.params.make_parameters_file

    Notes
    -----
    All parameters of this function are optional. Calling the function without
    providing any parameters will always produce the parameters file for the
    default model scenario. This file can be instantiated as a Model object and
    run without being edited. Those three steps (create default parameters file;
    create model from that parameters file; run the model) serve as a base case
    to test successful package installation, and are wrapped around by the
    convenience function `gnx.run_default_model`.

    Examples
    --------
    In the simplest example, we can create a parameters file for the default
    model. Then (assuming it is the only Geonomics parameters file in the
    current working directory, so that it can be unambiguously identified) we
    can call the gnx.make_model function to create a Model object from that
    file, and then call the Model.run method to run the model (setting the
    'verbose' parameter to True, so that we can observe model output).

    >>> gnx.make_parameters_file()
    >>> mod = gnx.make_model()

    NOTE: Using the following file, in the current working directory to create the Model object:
            GNX_params_21-01-2020_17:22:08.py

    >>> mod.run(verbose = True)
    ###############################################################################################
    Running model "GNX_params_21-01-2020_17:22:08"...
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Setting up iteration 0...
    Creating the burn-in function queue...
    Creating the main function queue...
    Running burn-in, iteration 0...
    burn:   it=0:   t=0
            species: spp_0                         N=250    (births=35      deaths=214)
    .......................................................................................
    burn:   it=0:  t=1
            species: spp_0                         N=250    (births=35      deaths=214)
    .......................................................................................
    burn:   it=0:  t=2
            species: spp_0                         N=250    (births=35      deaths=214)
    .......................................................................................
    .
    .
    .


    We could also use some of this function's arguments, to create a parameters
    file for a model with 3 Layers and 1 Species (all with the default
    components for their sections of the parameters file) and with a section
    for parameterizing data collection.

    >>> gnx.make_parameters_file(layers = 3, data = True)

    As a more complex example that is likely to be similar to most use cases,
    we can create a parameters file for a model scenario with:
        - 2 Layers (one being an nlmpy Layer that will not change over model
          time, the other being a raster read in from a GIS file and being
          subject to change over model time);
        - 2 Species (the first having genomes, 2 Traits, and movement
          that is dictated by a _ConductanceSurface; the second not having
          genomes but having dispersal determined by a
          _ConductanceSurface, and undergoing
          demographic change)
        - data-collection;
        - stats-collection;
    We can save this to a file named "2-spp_2-trait_model.py" in our current
    working directory.

    >>> gnx.make_parameters_file(
    >>>     #list of 2 dicts, each containing the values for each Layer's
    >>>     #parameters section
    >>>     layers = [
    >>>         {'type': 'nlmpy'},                              #layer 1 
    >>>         {'type': 'gis',                                 #layer 2 
    >>>          'change': True}
    >>>         ],
    >>>     #list of 2 dicts, each containing the values for each Species's
    >>>     #parameters section
    >>>     species = [
    >>>         {'genomes': True,                               #spp 1
    >>>          'n_traits': 2,
    >>>          'movement': True,
    >>>          'movement_surface': True},
    >>>         {'genomes': False,                              #spp 2
    >>>          'movement': True,
    >>>          'dispersal_surface': True,
    >>>          'demographic_change': True}
    >>>         ],
    >>>     #arguments to the data and stats parameters
    >>>     data = True, stats = True,
    >>>     #destination to which to write the resulting parameter file
    >>>     filepath = '2-spp_2-trait_model.py')

    """

    #check if any keys in the layers or species dicts are abnormal, and
    #provide warning if so
    valid_l_keys = ['type', 'change']
    valid_p_keys = ['movement', 'movement_surface', 'dispersal_surface',
        'genomes', 'n_traits', 'custom_genomic_architecture',
        'demographic_change', 'parameter_change']
    if isinstance(layers, list):
        for n, lyr_dict in enumerate(layers):
            if False in [k in valid_l_keys for k in lyr_dict.keys()]:
                invalid_keys = ', '.join([k for k in lyr_dict.keys(
                    ) if k not in valid_l_keys])
                err_msg = ("One or more of the keys in dict number %i of "
                    "the 'layers' argument is not valid. Invalid values: "
                    "%s.") % (n, invalid_keys)
                print('')
                raise ValueError(err_msg)
    if isinstance(species, list):
        for n, spp_dict in enumerate(species):
            if False in [k in valid_p_keys for k in spp_dict.keys()]:
                invalid_keys = ', '.join([k for k in spp_dict.keys(
                    ) if k not in valid_p_keys])
                err_msg = ("One or more of the keys in dict number %i of "
                    "the 'species' argument is not valid. Invalid values: "
                    "%s.") % (n, invalid_keys)
                raise ValueError(err_msg)

    _make_params_file(filepath=filepath, layers=layers, species=species,
                      data=data, stats=stats)


#wrapper around params.read
def read_parameters_file(filepath):
    """
    Create a new ParametersDict object.

    Read the Geonomics parameters file saved at the location indicated by
    'filepath', check its validity (i.e. that all the Layers and Species
    parameterized in that file have been given distinct names), then use the
    file to instantiate a ParametersDict object.

    Parameters
    ----------
    filepath : str
        String indicating the location of the Geonomics parameters file that
        should be made into a ParametersDict object.

    Returns
    -------
    :class:`geonomics.ParametersDict`
        A dict of nested dicts, all of which have key-value pairs whose values
        can be accessed using typical dict notation or using dot notation.
        Can be fed into the geonomics.make_model function to create a
        Geonomics model.


    Raises
    ------
    AssertionError
        If either the Layers or the Species parameterized in the parameters
        file have not all been given distinct names

    See Also
    --------
    sim.params.read
    sim.params.ParametersDict

    Examples
    --------
    Read a parameters file called "null_model.py" (located in the current
    working directory).

    >>> gnx.read_parameters_file('null_model.py')
    <class 'sim.params.ParametersDict'>
    Model name:                                     null_model

    """

    #first read in file as a block of text
    with open(filepath, 'r') as f:
        txt = f.read()
    #find all the layer names and spp names
    lyr_names = re.findall('\S+(?=: \{\n\n\s+#-*#\n\s+#--- layer num\.)', txt)
    lyr_names = [re.sub("'", '"', n) for n in lyr_names]
    spp_names = re.findall('\S+(?=: \{\n\n\s+#-*#\n\s+#--- spp num\.)', txt)
    spp_names = [re.sub("'", '"', n) for n in spp_names]
    #get Counter objects of each
    lyr_name_cts = C(lyr_names)
    spp_name_cts = C(spp_names)
    #assert that each layer name is used only once
    assert set([*lyr_name_cts.values()]) == {1}, ("At least one of the "
        "Layer names provided in the parameters file appears to be used more "
        "than once. Violating names include: %s") % (';'.join([( "'%s', "
        "used %i times.") % (str(k),
                            v) for k, v in lyr_name_cts.items() if v>1]))
    #assert that each spp name is used only once
    assert set([*spp_name_cts.values()]) == {1}, ("At least one of the "
        "Species names provided in the parameters file appears to be used "
        "more than once. Violating names include: %s") % (';'.join([("'%s', "
            "used %i times.") % (
                str(k), v) for k, v in spp_name_cts.items() if v>1]))

    #break the file into sections for each species, then check that
    #trait names are only used once within each spp
    spp_sects = re.split('#spp name', txt)
    for sect in spp_sects:
        trt_names = re.findall('\S+(?=: \{\n\s+#trait-selection)', sect)
        if len(trt_names) > 0:
            trt_names = [re.sub("'", '"', n) for n in trt_names]
            trt_name_cts = C(trt_names)
            sect_spp_name = re.findall(
                        '\S+(?=: \{\n\n\s+#-*#\n\s+#--- spp num\.)', txt)[0]
            assert set([*trt_name_cts.values()]) == {1}, ("At least one of the"
                " Trait names provided in the parameters for Species "
                "%s appears to be used more than once. "
                "Violating names include: %s") % (sect_spp_name,
                ';'.join([("'%s', used %i times.") % (str(k),
                v) for k, v in trt_name_cts.items() if v>1]))



    #now read the file in as a ParametersDict object
    params = _read_params_file(filepath)
    return(params)


#function to make a regular dict into a ParametersDict object
def make_params_dict(params, model_name=None):
    """
    Create a ParametersDict object from a dict object.

    Use a plain Python dict object, and an optional model name, to create
    a new ParametersDict object.

    Parameters
    ----------

    params: {dict}
        The plain Python dict that is to be turned into a Geonomics
        ParametersDict object.

    model_name: {str}, optional, default: None
        The name to be assigned to the model. (If not provided, the model
        will be called 'unnamed_model'.)

    Returns
    -------
    :class:`geonomics.ParametersDict`
        A dict of nested dicts, all of which have key-value pairs whose values
        can be accessed using typical dict notation or using dot notation. If
        formatted correctly, can be fed into the geonomics.make_model function
        to create a Geonomics model.


    """
    params_dict=ParametersDict(params)
    if model_name is not None:
        params_dict.model['name'] = model_name
    else:
        params_dict.model['name'] = 'unnamed_model'
    return params_dict


#function to create a model from a ParametersDict object
def make_model(parameters=None, verbose=False):
    """
    Create a new Model object.

    Use either a ParametersDict object or the path to a valid Geonomics
    parameters file (whichever is provided to the 'parameters' argument) to
    create a new Model object.

    Parameters
    ----------
    parameters : {ParametersDict, dict, str}, optional, default: None
        The parameters to be used to make the Model object.

        If `parameters` is a ParametersDict object, the object will be used to
        make the Model.

        If `parameters` is a dict object, Geonomics will attempt to convert it
        to a ParametersDict object, then use that object to make the Model.

        If `parameters` is a string, Geonomics will call
        `gnx.read_parameters_file` to make a ParametersDict object, then use
        that object to make the Model.

        If `parameters` is None, or is not provided, then Geonomics will
        attempt to find a single parameters file in the current working
        directory with the filename "GNX_params_<...>.py", will use that
        file to make a ParametersDict object, then will use that object to
        make the Model.

    Returns
    -------
    :class:`geonomics.Model`
        An object of the Model class

    Raises
    ------
    ValueError
        If the `parameters` argument was not provided and a single, valid
        Geonomics parameters file could not be identified in the current
        working directory
    ValueError
        If the `parameters` arugment was given a string that does not point
        to a valid parameters file
    ValueError
        If the ParametersDict provided to the `parameters` argument, or created
        from the parameters file being used, cannot be successfully made into a
        Model

    See Also
    --------
    gnx.read_parameters_file
    sim.model.Model

    Examples
    --------
    Make a Model from a single, valid "GNX_params_<...>.py" file that can
    be found in the current working directory (such as a file that would be
    produced by calling gnx.make_parameters_file without any arguments).

    >>> gnx.make_model()
    <class 'sim.model.Model'>
    Model name:                                     GNX_params_13-10-2018_15:54:03
    Layers:                                         0: '0'
    Species:                                        0: '0'
    Number of iterations:                           1
    Number of burn-in timesteps (minimum):          30
    Number of main timesteps:                       100
    Geo-data collected:                             {}
    Gen-data collected:                             {}
    Stats collected:                                {}


    Make a Model from a file called 'null_model.py', in the current working
    directory.

    >>> gnx.make_model('null_model.py')
    <class 'sim.model.Model'>
    Model name:                                     null_model
    Layers:                                         0: 'tmp'
                                                    1: 'ppt'
    Species:                                        0: 'C. fasciata'
    Number of iterations:                           2500
    Number of burn-in timesteps (mininum):          100
    Number of main timesteps:                       1000
    Geo-data collected:                             {csv, geotiff}
    Gen-data collected:                             {vcf, fasta}
    Stats collected:                                {maf, ld, mean_fit, het, Nt}

    """

    if parameters is None:
        try:
            params_files = [f for f in os.listdir('.') if (
                f.startswith('GNX_params_')
                and os.path.splitext(f)[1] == '.py')]
            assert len(params_files) == 1, ("The 'parameters' argument was not"
                " provided, and it appears that the current working directory "
                "contains more than one 'GNX_params_<...>.py' file. "
                "Please run again, providing a valid value for the "
                "'parameters' argument.")
            parameters = params_files[0]
            print(("\n\nNOTE: Using the following file, in the current working"
              " directory to create the Model object:\n\t%s\n\n") % parameters)
        except Exception as e:
            raise ValueError(("The 'parameters' argument was not provided, "
                "and Geonomics could not identify a single "
                "'GNX_params_<...>.py' file in the current working "
                "directory from which to create the Model object. The "
                "following error was thrown: %s") % e)

    assert ( (type(parameters) is str and os.path.isfile(parameters))
        or str(type(
            parameters)) == "<class 'geonomics.sim.params.ParametersDict'>"),(
        "If the 'parameters' argument is provided, its value must be either a "
        "string pointing to a valid Geonomics parameters file or an object of "
        "the ParametersDict class. If it is not provided, the current working "
        "directory must contain a single 'GNX_params_<...>.py' file "
        "from which to create the Model object.")

    if type(parameters) is str:
        try:
            parameters = read_parameters_file(parameters)
        except Exception as e:
            traceback.print_exc(file = sys.stdout)
            raise ValueError(("Failed to read the parameters file at the "
                "filepath that was provided. The following error was raised: "
                "\n\t%s\n\n") % e)

    if type(parameters) is dict:
        try:
            parameters = make_params_dict(params, 'unnamed_model')
        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            raise ValueError(("A plain dict was provided as the value for "
                              "the parameters argument, but that dict "
                              "could not be converted to a Geonomics "
                              "ParametersDict object."))

    #elif isinstance(parameters, geonomics.sim.params.ParametersDict):
    else:
        pass
    try:
        name = parameters.model.name
        mod = Model(name, parameters, verbose=verbose)
        return(mod)
    except Exception as e:
            traceback.print_exc(file = sys.stdout)
            raise ValueError(("Failed to create a Model object from the "
                "ParametersDict object being used. "
                "The following error was raised: \n\t%s\n\n") % e)


# convenience function for creating a parameters-file for, instantiating, and
# running the default model, plotting the result, then returning the Model
# object
def run_default_model(delete_params_file=True, animate=False):
    """
    Run the default Geonomics model.

    This will create a parameters file for the default Geonomics model,
    read that in as a model, and run a single iteration of that model.

    Parameters
    ----------
    delete_params_file : bool, optional, default: True
        Whether or not to delete the parameters file used for this model
        after the model has finished running

    animate : bool, optional, default: False
        Whether or not plot the model's species as an animated Matplotlib
        plot while the model is running.

    Returns
    -------
    :class:`geonomics.Model`
        The Model object that was created, after it has run to completion.
        (If saved to a variable, it can be used for any number of additional
        runs using the Model.walk method. It can also be used to try out
        visualization methods, or to introspect its structure.)
    """


    # make the default params file
    filename = 'GNX_default_model_params.py'
    make_parameters_file(filename)
    # create the default model
    mod = make_model(parameters = filename)
    # run the default model in verbose mode
    mod.walk(T=10000, mode='burn', verbose = True, animate=animate)
    mod.walk(T=50, mode='main', verbose=True, animate=animate)
    # plot the results
    mod.plot(0,0,0)
    # get rid of the params file it created
    if delete_params_file:
        os.remove(os.path.join('.', filename))
    # return the model, in case someone wants to mess with/introspect it
    # afterwards
    return mod


# convenience function for creating a parameters-file for, instantiating, and
# running the default model, plotting the result, then returning the Model
# object
def run_demo(name, save_figs=False, time_it=False, **kwargs):
    """
    Run a Geonomics demo.

    This will run whichever of the Geonomics demos is stipulated by the *name*
    argument.

    Parameters
    ----------
    name : str 
        Can be one of the following values:

            - *'IBD IBE'*: Runs the IBD, IBE demo (example 1 from the methods
              paper).

            - *'simult select'*: Runs the simultaneous selection demo (example
              2 from the methods paper).

            - *'Yosemite'*: Runs the Yosemite demo (example 3 from the methods
              paper).

    save_figs : bool, optional, default: False
        If True, all figures that are produced will be saved to their default
        filenames in the current working directory.

    time_it : bool, optional, default: False
        If True, the run time for the main portion of the model (excluding
        calculation of data for figures and creation of figures, as much as
        possible) will be calculated, then displayed after the model has run
        to completion.

    Notes
    -----
    Some of the available demos come from the original Geonomics methods paper,
    Hart and Wang 2020, "Geonomics: forward-time, agent-based, spatially
    explicit, and arbitrarily complex landscape genomic simulations".

    Returns
    -------
    :class:`geonomics.Model`
        The Model object that was created, after it has run to completion.
        (If saved to a variable, it can be used for any number of additional
        runs using the Model.walk method. It can also be used to try out
        visualization methods, or to introspect its structure.)
    """

    if name.lower() == 'ibd ibe':
        params = demos._IBD_IBE._make_params()
        params = make_params_dict(params)
        mod = demos._IBD_IBE._run(params, save_figs, time_it, **kwargs)

    elif name.lower() == 'simult select':
        params = demos._simult_select._make_params()
        params = make_params_dict(params)
        mod = demos._simult_select._run(params, save_figs, time_it, **kwargs)

    elif name.lower() == 'yosemite':
        params = demos._yosemite._make_params()
        params = make_params_dict(params)
        mod = demos._yosemite._run(params, save_figs, time_it, **kwargs)

    else:
        print(('The specified demo ("%s") either is not yet implemented '
               'or does not exist!') % name)
        mod = None

    return mod


# wrapper around landscape.make_landscape
def make_landscape(params):
    """
    Make a Landscape object.

    Use the parameter values stored in the input params argument
    to instantiate a Landscape object.

    Parameters
    ----------
    params : ParametersDict
        A ParametersDict object, such as that output by the
        make_params_dict function.

    Returns
    -------
    :class:`geonomics.structs.landscape.Landscape`
        A Landscape instance.
    """

    land = _make_landscape(params)
    return land


#wrapper around genome.make_genomic_architecture
def make_genomic_architecture(params, landscape):
    """
    Make a GenomicArchitecture object.

    Use the parameter values stored in the input params argument,
    and the Landscape object input as the landscape argument,
    to instantiate a GenomicArchitecture object.

    Parameters
    ----------
    params : ParametersDict
        A ParametersDict object (such as that output by the
        make_params_dict function).

    landscape : Landscape
        A Landscape object (such as that returned by the
        make_landscape function).

    Returns
    -------
    :class:`geonomics.structs.genome.GenomicArchitecture`
        A GenomicArchitecture instance.

    """
    gen_arch = _make_genomic_architecture(params, landscape)
    return gen_arch


#wrapper around individual.make_individual
    #should provide either a new genome for the individual, or a
    #genomic_architecture to use to draw its genome;
    #and should provide either a parental centerpoint to disperse from, or a
    #landscape.dim tuple within which to choose a location;
    #burn can be True (i.e. then the individual will have a [[0,0]] genome)
def make_individual(idx, genomic_architecture=None, new_genome=None, dim=None,
    parental_midpoint=None, sex=None, age=0):
    """
    Make an Individual object.

    Make a new Individual, whose index number, genome,
    location, sex, and age are determined by the input arguments.

    Parameters
    ----------
    idx : int
        The index number to be assigned to the new individual.

    genomic_architecture : GenomicArchitecture, optional
        A GenomicArchitecture object, describing the genomic
        characteristics of the Species to which the Individual belongs.
        If the new_genome argument is not provided then the relevant parameter
        values in the GenomicArchitecture will be used when randomly
        drawing the individual's new genome.

    new_genome : nx2 np.ndarray, optional
        The Individual's genome, as an nx2 Numpy ndarray of 0's and 1's.

    dim : (int, int), optional
        A tuple of the landscape's (x, y) dimensionality, to be used if the
        individual's location is to be randomly randomly drawn from anywhere on
        the landscape (rather than being drawn by dispersal from the parental
        midpoint, which will occur if the parental_midpoint argument is
        provided).

    parental_midpoint : (float, float), optional
        A tuple of the (x,y) continuous coordinates, representing the point
        from which dispersal should occur to determine the individual's
        location. This is normally thought of as the midpoint between the individual's
        two parents, but if you are creating an individual without reference to
        parents then this can also simply be thought of as representing the center of
        the neighborhood into which the individual will disperse.

    sex : int, optional
        An integer representing the individual's sex (0 for female, 1 for
        male). Can be left as None for asexual or non-sexed Species.

    age : int, optional
        An integer representing the number of timesteps for which the
        individual has already survived - hence its age, or stage.

    Returns
    -------
    :class:`geonomics.structs.individual.Individual`
        An Individual instance.
    """
    assert (genomic_architecture is not None
        or new_genome is not None), ("Either a new genome must be provided "
        "(i.e. 'new_genome' must not be None) or a genomic architecture from "
        "which to draw a new genome must be provided (i.e. "
        "'genomic_architecture' must not be None.")
    assert (parental_centerpoint is not None
            or dim is not None), ("Either a Landscape-dimension tuple must be "
            "provided (i.e. 'dim' must not be None) or a parental centerpoint "
            "from which to disperse the individual must be provided (i.e. "
            "'parental_centerpoint' must not be None).")
    ind = _make_individual(idx = idx, offspring = False,
            dim = dim, genomic_architecture = genomic_architecture,
            new_genome = new_genome, sex = sex,
            parental_centerpoint = parental_centerpoint, age = age)
    return ind


#wrapper around species.make_species
    #burn can be True (i.e. then the individuals will have a [[0,0]] genome)
def make_species(landscape, spp_params):
    """
    Make a Species object.

    Use the parameter values stored in the input spp_params argument,
    and the Landscape object input as the landscape argument,
    to instantiate a Species object.

    Parameters
    ----------
    landscape : Landscape
        A Landscape object (such as that returned by the
        make_landscape function).

    spp_params : ParametersDict
        The subsection of a full ParametersDict object (such as that output by the
        make_params_dict function) that pertains to the particular species to
        be created.

    Returns
    -------
    :class:`geonomics.structs.species.Species`
        A Species instance.
    """
    spp = _make_species(landscape, spp_params)
    return(spp)


#wrapper around community.make_comunity
    #burn can be True (i.e. then the individuals will have a [[0,0]] genome)
def make_community(landscape, params):
    """
    Make a Community object.

    Use the parameter values stored in the input params argument,
    and the Landscape object input as the landscape argument,
    to instantiate a Community object.

    Parameters
    ----------
    landscape : Landscape
        A Landscape object (such as that returned by the
        make_landscape function).

    spp_params : ParametersDict
        A ParametersDict object (such as that output by the
        make_params_dict function).

    Returns
    -------
    :class:`geonomics.structs.community.Community`
        A Community instance.
    """
    comm = _make_community(landscape, params)
    return comm



#!/usr/bin/python
# landscape.py

'''
Defines the Layer and Landscape classes, with their associated methods and
supporting functions
'''
#geonomics imports
from geonomics.utils.viz import _get_plt_lims, _plot_rasters, _check_display
from geonomics.utils.io import _write_geotiff, _write_txt_array, _read_raster
from geonomics.utils.spatial import _scale_raster, _make_nlmpy_raster
from geonomics.ops.change import _LandscapeChanger

#other imports
import numpy as np
import numpy.random as r
import random
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from collections import Counter as C
from copy import deepcopy
from operator import itemgetter as ig
from scipy import interpolate
from shapely import geometry as g


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################

class Layer:
    """
    Representation of a single environmental layer (i.e. variable).

    Multiple Layers are collected as serial integer-keyed values within
    a Landscape dict.

    Attributes
    ----------

        NOTE: For more detail, see the documentation for the parameters
              that correspond to many of the following attributes.

        coord_prec:
            The precision (i.e. number of significant digits) to which
            coordinate values should be round when the Layer is plotted.
       
        dim:
            The x,y (i.e. lon,lat; or i,j in array terms) dimensions of the
            Layer. (Must be the same as the Landscape to which the Layer
            belongs.)

        idx:
            Index number of the Layer (i.e. its key within the Landscape dict)

        name:
            The string name of the Layer

        prj:
            The projection of the Layer (formatted as a PROJ4 string).
            (Must be the same as the Landscape to which the Layer belongs.)

        rast:
            The 2d numpy array, of shape `Landscape.dim`, containing the
            environmental values for this Layer.

        res:
            The x,y (i.e. lon,lat; or i,j in array terms) spatial resolution
            (i.e. cell sizes) of the layer.  (Must be the same as the
            Landscape to which the Layer belongs.)


        type:
            A string indicating the type of the Layer ('random', 'defined',
            'file', or 'nlmpy')

        ulc: 
            The x,y (i.e. lon,lat; or i,j in array terms) coordinates of the
            upper left corner of the layer. (Must be the same as the Landscape
            to which the Layer belongs.)

        units:
            A string representation of the units of the Layer's variable
            (to be used for plotting).
    
    """

    #######################
    ### SPECIAL METHODS ###
    #######################

    def __init__(self, rast, lyr_type, name, dim, res=(1,1), ulc=(0,0),
                 prj=None, coord_prec=0, units='', scale_min=0, scale_max=1):
        self.idx = None
        self.type = lyr_type
        self.name = str(name)
        #the x,y (i.e. j,i) dimensions
        self.dim = dim
        assert type(self.dim) in [tuple, list], ("dim must be expressed "
                                                 "as a tuple or a list")
        #set the resoultion (res; i.e. x,y cell sizes) and upper-left corner
        #(ulc) to defaults; will be reset if layer read in from a GIS raster or
        #numpy txt array file
        self.res = res
        self.ulc = ulc
        self.prj = prj
        self.coord_prec = coord_prec
        self.units = units
        # attributes for arrays of cell-bounds coordinates, to be set when a
        # Landscape is instantiated
        self._x_cell_bds = None
        self._y_cell_bds = None
        self.rast = rast
        assert type(self.rast) == np.ndarray, "rast should be a numpy.ndarray"
        self._scale_min = scale_min
        self._scale_max = scale_max
        #attribute that will track the Species for which this Layer
        #is the K_layer (to be used to update
        #Species.K if this Layer undergoes any landscape changes)
        self._is_K = []


    #####################
    ### OTHER METHODS ###
    #####################

    #################
    #private methods#
    #################

    #method for writing the layer's raster to a geotiff raster file
    def _write_geotiff(self, filepath):
        _write_geotiff(filepath, self)

    #method for writing the layer's raster to a numpy txt array
    def _write_txt_array(self, filepath):
        _write_txt_array(filepath, self)

    # method for the getting the vmin and vmax values for plotting
    def _get_plot_vmin_vmax(self):
        return (self._scale_min, self._scale_max)

    #method for plotting the layer
    def _plot(self, cbar=True, cmap=None, x=None, y=None, zoom_width=None,
              vmin=None, vmax=None, ticks=None, mask_rast=None):
        plt_lims = _get_plt_lims(self, x, y, zoom_width)
        # get the vmin and vmax values, if not provided
        if vmin is None and vmax is None:
            vmin, vmax = self._get_plot_vmin_vmax()
        _plot_rasters(self, cbar=cbar, cmap=cmap,
                      plt_lims=plt_lims, vmin=vmin, vmax=vmax, ticks=ticks,
                      mask_rast=mask_rast)

    # method to get the tickmarks to use to plot the layer
    def _get_coord_ticks(self):
        x_ticks = np.int64(np.round(np.linspace(0, self.dim[0]-1, 8)))
        y_ticks = np.int64(np.round(np.linspace(0, self.dim[1]-1, 8)))
        x_tick_labs = [round(self.ulc[0] + (self.res[0] * (loc + 0.5)),
                             min(3, self.coord_prec)) for loc in x_ticks]
        y_tick_labs = [round(self.ulc[1] + (self.res[1] * (loc + 0.5)),
                             min(3, self.coord_prec)) for loc in y_ticks]
        return x_ticks, x_tick_labs, y_ticks, y_tick_labs

    # method for recovering a file-based raster's values expressed in its
    # native units (i.e. undoing the 0-1 scaling)
    def _get_rast_in_native_units(self):
        native = (self.rast * (self._scale_max
                               - self._scale_min)) + self._scale_min
        return native

    # method for getting colorbar ticks
    def _get_cbar_ticks_and_minmax_scaled_vals(self):
        #if self.type == 'file':
            #rast = self._get_rast_in_native_units()
            #min_val = rast.min()
            #max_val = rast.max()
        #else:
        #    min_val = self.rast.min()
        #    max_val = self.rast.max()
        min_val = self._scale_min
        max_val = self._scale_max
        ticks = np.round(np.linspace(min_val, max_val, 5), 3)
        return ticks, min_val, max_val


    #method for writing the lyr's raster to a file of the specified format
    def _write_raster(self, filepath, raster_format):
        assert raster_format in ['geotiff', 'txt'], ("The raster_format "
            "must be one of the following: 'geotiff', 'txt'.")
        if raster_format == 'geotiff':
            self._write_geotiff(filepath)
        elif raster_format == 'txt':
            self._write_txt_array(filepath)


class Landscape(dict):
    """
    Representation of a multi-layer (i.e. multivariate) landscape.

    Organized as a dict of multiple, serial integer-keyed Layer objects,
    
    Because the Landscape class inherits from `dict`, Layers can
    be indexed out using their index-number keys (e.g. `mod.land[<idx>]`).

    The Landscape is stored as the 'mod.land' attribute of its corresponding
    Model object.

    Attributes
    ----------

        NOTE: For more detail, see the documentation for the parameters
              that correspond to many of the following attributes.

        dim:
            The x,y (i.e. lon,lat; or i,j in array terms) dimensions of the
            Layer. (Must be the same as the Landscape to which the Layer
            belongs.)

        n_lyrs:
            The number of Layers in the Landscape

        prj:
            The projection of the Layer (formatted as a PROJ4 string).
            (Must be the same as the Landscape to which the Layer belongs.)

        res:
            The x,y (i.e. lon,lat; or i,j in array terms) spatial resolution
            (i.e. cell sizes) of the layer.  (Must be the same as the
            Landscape to which the Layer belongs.)

        ulc: 
            The x,y (i.e. lon,lat; or i,j in array terms) coordinates of the
            upper left corner of the layer. (Must be the same as the Landscape
            to which the Layer belongs.)
            
    """

    #######################
    ### SPECIAL METHODS ###
    #######################

    def __init__(self, lyrs, res=(1,1), ulc=(0,0), prj=None, mod=None):
        #check the lyrs dict is correct, then update the Landscape with it
        assert False not in [
            lyr.__class__.__name__ == 'Layer' for lyr in lyrs.values(
            )], ('All layers supplied in lyrs must be of type '
            'landscape.Layer.')
        self.update(lyrs)
        #set the lyrs attribute to the dict values (it still changes
        #dynamically as things update, it runs
        #3x faster than calling values() during the model, and it makes
        #code nicer looking
        #FIXME: I had been using the following lines, but deepcopy breaks
        #because it cannot pickle dict.values() objects
        #self.lyrs = self.values()
        #set the number of lyrs in the stack
        self.n_lyrs = len(self)
        #set the idx attributes on the lyrs
        [setattr(v, 'idx', k) for k, v in self.items()]

        #check that lyrs' dimensions are all the same
        assert len(set([land.dim for land in list(self.values(
            ))])) == 1, 'Dimensions of all layers must be equal.'
        #then set the dim to the 0th layer's dim 
        self.dim = list(self.values())[0].dim
        #get the order of magnitude of the larger land dimension
        #(used when zero-padding cell-coorindate strings)
        self._dim_om = max([len(str(d)) for d in self.dim])
        #set the resoultion (res; i.e. x,y cell sizes), upper-left corner
        #(ulc), and projection (prj) to the provided values
        self.res = res
        # get the resolution ratio (for calculating individuals' movement
        # distances on non-square-resolution rasters)
        self._res_ratio = tuple(np.abs([val/max(self.res) for val in
                                        self.res]))
        self.ulc = ulc
        self.prj = prj
        # get arrays of grid-cell x and y bounds , for plotting
        # with plt.pcolormesh
        x_cell_bds, y_cell_bds = [np.linspace(self.ulc[i],
                                              self.ulc[i] + (self.res[i] * (
                                                             self.dim[i])),
                                              self.dim[i] + 1) for i in range(
                                                                            2)]
        self._x_cell_bds = x_cell_bds
        self._y_cell_bds = y_cell_bds
        #And check that the Layers' res, ulc, and prj values are equal to 
        #the Landscape's values
        assert np.all([lyr.res == self.res for lyr in self.values(
            )]), ("Not all Layers have the same resolution value "
            "(attribute 'res') as the 'res' value being used to create "
            "the Landscape object that should contain them.")
        assert np.all([lyr.ulc == self.ulc for lyr in self.values(
            )]), ("Not all Layers have the same upper-left corner value "
            "(attribute 'ulc') as the 'ulc' value being used to create "
            "the Landscape object that should contain them.")
        assert np.all([lyr.prj == self.prj for lyr in self.values(
            )]), ("Not all Layers have the same projection (attribute 'prj') "
            "as the 'prj' value being used to create the Landscape "
            "object that should contain them.")
        #check that all layers' values are constrained to 0 <= x <= 1
        for lyr in lyrs.values():
            assert np.all(0 <= lyr.rast), ("Layer '%s' contains values "
                                           "less than 0.") %(lyr.name)
            assert np.all(lyr.rast <= 1), ("Layer '%s' contains values "
                                           "greater than 1.") %(lyr.name)


        #create a changer attribute (defaults to None, but will later be set
        #to an ops.change.LandscapeChanger object if params call for it)
        self._changer = None

    #define the __str__ and __repr__ special methods
    #NOTE: this doesn't excellently fit the Python docs' specification for 
    #__repr__; I should massage this some more when done writing the codebase
    def __str__(self):
        #get the longest lyr name, to be used to horizontally rectify 
        #all the lines for each of the lyrs
        max_len_lyr_name = max([len(lyr.name) for lyr in self.values()])
        #get a string representation of the class
        type_str = str(type(self))
        #get a string of all the lyrs
        lyrs_str = '%i Layer%s:\n' % (self.n_lyrs, 's' * (len(self) > 1))
        lyrs_str = lyrs_str + '\n'.join(['\tlyr %i: ' % k +
            "%s'%s':\t" % (' ' * (max_len_lyr_name - len(v.name)), v.name) +
            str(v) for k,v in self.items()])
        #get a string representation of the first two and last two parameters
        #params_str = "\nParameters:\n\t" + ',\n\t'.join(sorted([str(k) +
        #    ': ' + str(v) for k,v in vars(self).items()][:2])) + ','
        #params_str = params_str + '\n\t...\n\t'
        #params_str = params_str + ',\n\t'.join(sorted([str(k) +
        #    ': ' + str(v) for k,v in vars(self).items()][-2:]))
        #return '\n'.join([type_str, lyrs_str, params_str])
        return '\n'.join([type_str, lyrs_str])

    def __repr__(self):
        repr_str = self.__str__()
        return repr_str


    #####################
    ### OTHER METHODS ###
    #####################

        #################
        #private methods#
        #################

    #method to set a raster
    def _set_raster(self, lyr_num, rast):
        self[lyr_num].rast = rast

    #method to set all lyrs' res and ulc attributes
    def _set_lyrs_res_ulc_prj(self):
        [setattr(lyr, 'res', self.res) for lyr in self.values()]
        [setattr(lyr, 'ulc', self.ulc) for lyr in self.values()]
        [setattr(lyr, 'prj', self.prj) for lyr in self.values()]

    #method to make landscape changes
    def _make_change(self, t, verbose=False):
        self._changer._make_change(t=t, additional_args={'land': self},
                                   verbose=verbose)

    #method to plot the landscape (or just a certain lyr)
    def _plot(self, lyr_num=None, cbar=True, cmap=None, x=None, y=None,
              zoom_width=None, vmin=None, vmax=None, ticks=None,
              mask_rast=None):
        plt_lims = _get_plt_lims(self, x, y, zoom_width)
        # get the vmin and vmax values, if not provided
        if vmin is None and vmax is None:
            if lyr_num is None:
                vmins_vmaxs = [lyr._get_plot_vmin_vmax(
                                                    ) for lyr in self.values()]
                vmin = [v[0] for v in vmins_vmaxs]
                vmax = [v[1] for v in vmins_vmaxs]
            else:
                vmin, vmax = self[lyr_num]._get_plot_vmin_vmax()
        _plot_rasters(self, lyr_num=lyr_num, cbar=cbar, cmap=cmap,
                      plt_lims=plt_lims, vmin=vmin, vmax=vmax, ticks=ticks,
                      mask_rast=mask_rast)

    # method to get the tickmarks to use to plot the layer
    def _get_coord_ticks(self):
        try:
            coord_prec = min([lyr.coord_prec for lyr in self.values(
                            ) if lyr.coord_prec != 0])
        except ValueError:
            coord_prec = 0
        x_ticks = np.int64(np.round(np.linspace(0, self.dim[0]-1, 8)))
        y_ticks = np.int64(np.round(np.linspace(0, self.dim[1]-1, 8)))
        x_tick_labs = [round(self.ulc[0] + (self.res[0] * (loc + 0.5)),
                             min(3, coord_prec)) for loc in x_ticks]
        y_tick_labs = [round(self.ulc[1] + (self.res[1] * (loc + 0.5)),
                             min(3, coord_prec)) for loc in y_ticks]
        return x_ticks, x_tick_labs, y_ticks, y_tick_labs

        ################
        #public methods#
        ################

    # method for pickling a landscape
    def _write_pickle(self, filename):
        import cPickle
        with open(filename, 'wb') as f:
            cPickle.dump(self, f)


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

def _make_random_lyr(dim, n_pts, interp_method="cubic", num_hab_types=2,
                       dist='beta', alpha=0.05, beta=0.05):
    # requires layer to be square, such that dim = domain = range

    # NOTE: can use "nearest" interpolation to create random patches of
    #habitat (by integer values); can change num_hab_types to > 2
    #to create a randomly multi-classed layer
    # NOTE: I guess this could be used for rectangular layers, if the
    #square raster is generated using the larger of the two dimensions,
    #and then the resulting array is subsetted to the layer's dimensions
    # n_pts/dim ratio:
    # ~0.01-0.05 --> broad, simple gradients
    # ~0.05-0.10 --> slightly more complex, occasionally layer of
    #low regions divided by a major high region (or vice versa)
    # ~0.10-0.50 --> layer can be broken up into numerous fragmented
    #patches (though sometimes still relatively homogeneous, with
    #one or two small, extremely different patches
    # ~0.50-1.00 --> more highly fragmented version of above
    max_dim = max(dim)
    if interp_method == 'nearest':
        if dist == 'unif':
            vals = r.rand(n_pts) * (num_hab_types - 1)
        elif dist == 'beta':
            vals = r.beta(alpha, beta, n_pts) * (num_hab_types - 1)
    else:
        if dist == 'unif':
            vals = r.rand(n_pts)
        elif dist == 'beta':
            vals = r.beta(alpha, beta, n_pts)
    # select seed points from well outside the eventaul layer, to ensure
    #interpolation across area of interest
    pts = r.normal(max_dim / 2, max_dim * 2, [n_pts,
                                              2])
    grid_x, grid_y = np.mgrid[1:max_dim:complex("%ij" % max_dim),
                            1:max_dim:complex( "%ij" % max_dim)]
        #NOTE: by this function's definition, the use of complex numbers
        #in here specifies the number of steps desired
    I = interpolate.griddata(pts, vals, (grid_x, grid_y), method=interp_method)
    if interp_method == 'nearest':
    # i.e., if being used to generate random habitat patches...
        I = I.round().astype(float)
    if interp_method == 'cubic':
        # transform to constrain all values to 0 <= val <= 1
        I = I + abs(I.min()) + (
        # NOTE: adding a small jitter to keep values from reaching == 0
        #or == 1, as would likely be the case with linear interpolation
                0.01 * r.rand())
        I = I / (I.max() + (0.01 * r.rand()))
    #use the dim tuple to subset an approriate size if dim not equal
    if dim[0] != dim[1]:
        I = I[:dim[0], :dim[1]]
    return I


def _make_defined_lyr(dim, rast, pts, vals, interp_method="cubic",
                                                    num_hab_types=2):
    #pts should be provided as n-by-2 Numpy array, vals as a 1-by-n Numpy array

    # NOTE: There seem to be some serious issues with all of this code,
    #because the resulting layers are not quite symmetrical;
    #and making small tweaks (e.g. adding 5 to all input points'
    #coordinates) doesn't just tweak the output layer but instead
    #completely gets ride of it; I have an intuition that it comes
    #from the code that coerces all raster values to 0 <= val <= 1,
    #becuase that doesn't look it does quite what I originally intended for
    #it to do, but I'm not really sure... anyhow, for now it works for
    #my initial testing purposes

    # NOTE: like the _make_random_lyr function, this also 
    #requires layer to be square, but I guess this could be
    #used for rectangular layers, if the square raster is
    #generated using the larger of the two dimensions, and
    #then the resulting array is subsetted to the layer's dimensions

    # NOTE: if discrete habitat patches desired, values should still
    #be fed in as proportions, and will then be multipled by
    #num_hab_types to develop habitat class values
    if rast is not None:
        I = rast
    else:
        if interp_method == 'nearest':
            vals = vals * (num_hab_types - 1)

        max_dim = max(dim)
        grid_x, grid_y = np.mgrid[1:max_dim:complex("%ij" % max_dim),
                                  1:max_dim:complex("%ij" % max_dim)]
        I = interpolate.griddata(pts, vals, (grid_x, grid_y), method=interp_method)
        if interp_method == 'nearest':
        # i.e., if being used to generate random habitat patches...
            I = I.round().astype(float)
        if interp_method == 'cubic':
            # transform to constrain all values to 0 <= val <= 1
            I = I + abs(I.min()) + (
                    0.01 * r.rand())
                # NOTE: adding a small jitter to keep values from
                #reaching == 0 or == 1,
            # as would likely be the case with linear interpolation
            I = I / (I.max() + (0.01 * r.rand()))
        #use the dim tuple to subset an approriate size if dim not equal
        if dim[0] != dim[1]:
            I = I[:dim[0], :dim[1]]
    return I


def _make_landscape(mod, params, num_hab_types=2, verbose=False):
    #print verbose output
    if verbose:
        print('\tMAKING LANDSCAPE...\n')
    # NOTE: If a multi-class (rather than binary) block-habitat raster
    #would be of interest, would need to make
    # num_hab_types customizable)
    dim = params.landscape.main.dim
    res = params.landscape.main.res
    ulc = params.landscape.main.ulc
    prj = params.landscape.main.prj
    if res is None:
        res = (1,1)
    if ulc is None:
        ulc = (0,0)
    #leave default projection as None for now
    if prj is None:
        prj = None

    #create a dictionary to hold all the lyrs to be created
    lyrs = {}

    #create a list to hold the information for any rasters to be created from
    #file (because they will all be created after the loop over the lyrs, so
    #that it's simpler to check agreement among raster resolutions and 
    #registrations
    file_lyr_params = {'names': [],
                       'lyr_nums': [],
                       'filepaths': [],
                       'scale_min_vals': [],
                       'scale_max_vals': [],
                       'coord_precs': [],
                       'unitss': []}

    #then loop over the lyrs in params.landscape.lyrs and create each one
    for n, (lyr_name, lyr_params) in enumerate(
                                        params.landscape.layers.items()):

        #get the init parameters
        init_params = deepcopy(lyr_params.init)

        #determine which type of lyr this is to be
        #(valid: 'random', 'defined', 'file')
        init_keys = [*init_params]
        if len(init_keys) > 1:
            raise ValueError(("The %ith layer (params['land']['layers'] "
                "key '%s') appears to have parameters for more than one layer "
                "type.  Choose a single layer type (valid values: 'random', "
                "'defined', 'file', 'nlmpy') and provide a sub-dictionary "
                "of parameters for only that type.") % (n, str(lyr_name)))
        lyr_type = init_keys[0]
        assert lyr_type in ['random', 'defined', 'file', 'nlmpy'], ("The "
            "parameters sub-dictionary for the %ith layer (params['land']"
            "['layers'] key '%s') has an invalid key value. Valid keys are: "
            "'random', 'defined', 'file'.") % (n, str(lyr_name))

        #create a random lyr, if called for
        if lyr_type == 'random':
            #create the random lyr and add it to the lyrs dict
            lyr_rast = _make_random_lyr(dim, **init_params[lyr_type],
                                           num_hab_types=num_hab_types)
            lyrs[n] = Layer(lyr_rast, lyr_type = lyr_type,
                name = lyr_name, dim = dim, res = res, ulc = ulc)

        #or else create a defined lyr 
        elif lyr_type == 'defined':
            #create the defined raster
            lyr_rast = _make_defined_lyr(dim, **init_params[lyr_type],
                                            num_hab_types=num_hab_types)
            lyrs[n] = Layer(lyr_rast, lyr_type = lyr_type,
                name = lyr_name, dim = dim, res = res, ulc = ulc)

        #or else create an nlmpy lyr
        elif lyr_type == 'nlmpy':
            #get the params
            nlmpy_params = init_params[lyr_type]
            #make the nlm
            lyr_rast = _make_nlmpy_raster(nlmpy_params)
            #check that its dimensions match those of the Landscape
            assert lyr_rast.shape == dim, ("The dimensions of the NLM "
                "created appear to differ from the Landscape dims: "
                "Landscape has x,y (i.e. j,i) dims %s, "
                "NLM has dims %s.\n\n") % (str(dim), str(lyr_rast.shape))
            #set the nlm as the nth Layer in the Landscape
            lyrs[n] = Layer(lyr_rast, lyr_type = lyr_type,
                name = lyr_name, dim = dim, res = res, ulc = ulc)

        #or else create a lyr from file
        elif lyr_type == 'file':
            #set this lyr to None, temporarily
            lyrs[n] = None
            #then store the relevant info for this lyr in the 
            #file_lyr_params dict; this lyr's raster
            #will be replaced with a GIS or numpy raster after the Landscape is 
            #created (NOTE: easier to do this all at 
            #once afterward, to check that the resolution and registration
            #of the rasters all agree)
            file_lyr_params['lyr_nums'].append(n)
            file_lyr_params['names'].append(lyr_name)
            [file_lyr_params[k+'s'].append(v) for k,v in init_params[
                                                        lyr_type].items()]

    #now set the necessary layers to their file rasters, if applicable
    if True in [len(v) > 0 for v in file_lyr_params.values()]:
        file_lyrs, res, ulc, prj, = _get_file_rasters(land_dim=dim,
                                                      **file_lyr_params)
        for n in file_lyr_params['lyr_nums']:
            lyrs[n] = file_lyrs[n]

        #set all other layers in the lyrs dict to have the same res, ulc,
        #and prj (because they all were not read in from file, whereas all
        #file-based layers were, and their res, ulc, and prj values were
        #already checked for equality, so we make the strong assumption
        #that the user wants the remaining non-file layers to be represented
        #by the same res, ulc, and prj values)
        for lyr_name in [*lyrs]:
            lyrs[lyr_name].res = res
            lyrs[lyr_name].ulc = ulc
            lyrs[lyr_name].prj = prj
        [setattr(lyrs[n], 'coord_prec',
                 v) for n, v in zip(file_lyr_params['lyr_nums'],
                                    file_lyr_params['coord_precs'])]

    #create the land object
    land = Landscape(lyrs, res=res, ulc=ulc, prj=prj, mod=mod)

    # set all Layers' cell-bounds attributes
    [setattr(lyr, '_x_cell_bds', land._x_cell_bds) for lyr in land.values()]
    [setattr(lyr, '_y_cell_bds', land._y_cell_bds) for lyr in land.values()]

    #grab the change parameters into a dictionary of
    #lyr_num:events:events_params hierarchical
    change_params = {k:v.change for k,v in params.landscape.layers.items(
                                                ) if 'change' in v.keys()}
    #and then replace the land._changer attribute with
    #an ops.change.LandscapeChanger object, if params call for it
    if len(change_params) > 0:
        #replace the keys, which are layer names, with layer numbers
        lyr_num_change_params = {}
        for k,v in change_params.items():
            lyr_num = [num for num, lyr in land.items() if lyr.name == k]
            assert len(lyr_num) == 1, ("Expected to find a single Layer "
                "matching the layer names for each Landscape-change event, "
                "but found %i Layers matching Layer name "
                "'%s'.") % (len(lyr_num), k)
            lyr_num_change_params[lyr_num[0]] = v
        land._changer = _LandscapeChanger(land, lyr_num_change_params,
            mod = mod)

    return land


def _get_file_rasters(land_dim, names, lyr_nums, filepaths, coord_precs,
                      scale_min_vals, scale_max_vals, unitss):
    assert len(lyr_nums)==len(filepaths), ('Parameters provide a '
                                           'different number of GIS raster '
                                           'files to read in than of layer '
                                           'numbers to set them to.')
    res = []
    ulc = []
    prj = []
    rasters = []
    lyrs = []
    for n,filepath in enumerate(filepaths):
        #get array, dim, res, ulc, and prj from io.read_raster
        rast_array, rast_dim, rast_res, rast_ulc, rast_prj = _read_raster(
            filepath, coord_precs[n], land_dim)
        #check that the dimensions are right
        assert rast_dim == land_dim, ('Variable land_dim and the dimensions '
            'of the input raster %s appear to differ. Please clip %s to '
            'the correct dimensions and try again, because it has x,y '
            'dimensions (%i, %i), but land has x,y '
            'dimensions (%i,%i)') % (filepath,
            filepath, rast_dim[0], rast_dim[1], land_dim[0], land_dim[1])
        #if either the scale_min_val or scale_max_val for this Layer is None,
        #set it to the min or max value of this Layer's array
        if scale_min_vals[n] is None:
            scale_min_vals[n] == rast_array.min()
        if scale_max_vals[n] is None:
            scale_max_vals[n] == rast_array.max()
        #scale the raster to 0<=x<=1
        rast_array, scale_min, scale_max = _scale_raster(rast_array,
                min_inval = scale_min_vals[n], max_inval = scale_max_vals[n])
        #add the rast_array to the rast_arrays list
        rasters.append(rast_array)
        #and add the raster's ulc and res attributes to the appropriate lists
        res.append(tuple(rast_res))
        ulc.append(tuple(rast_ulc))
        prj.append(rast_prj)
        #and update scale_min_vals and scale_max_vals if the values used
        #in spt.scale_raster() don't equal the values from params
        #(should only occur if the params values provided were None)
        if scale_min != scale_min_vals[n]:
            scale_min_vals[n] = scale_min
        if scale_max != scale_max_vals[n]:
            scale_max_vals[n] = scale_max
    #check that there is only one unique tuple each in res and ulc lists,
    #and only one unique WKT string in the prj list
    #(otherwise rasters are not in the same projection and/or registration)
    res_C = C(res)
    ulc_C = C(ulc)
    prj_C = C(prj)
    assert len(res_C) == 1 and list(res_C.values())[0] == len(res), ('Some of '
        'the rasters read in from file appear to differ in resolution.')
    assert len(ulc_C) == 1 and list(ulc_C.values())[0] == len(ulc), ('Some of '
        'the rasters read in from file appear to differ in registration (i.e. '
        'top-left corner).')
    assert len(prj_C) == 1 and list(prj_C.values())[0] == len(prj), ('Some of '
        'the rasters read in from file appear to have different projections.')
    #then get the res and ulc values
    res = res[0]
    ulc = ulc[0]
    prj = prj[0]
    #create lyrs from the rasters
    lyrs = [Layer(rast, lyr_type='file', name=name, dim=land_dim, res=res,
                  ulc=ulc, prj=prj, scale_min=scale_min, scale_max=scale_max,
                 units=units) for lyr_num, name,
            rast, scale_min, scale_max, units in zip(lyr_nums, names, rasters,
            scale_min_vals, scale_max_vals, unitss)]
    return(lyrs, res, ulc, prj)


# function for reading in a pickled landscape stack
def read_pickled_land(filename):
    import cPickle
    with open(filename, 'rb') as f:
        land = cPickle.load(f)
    return land

